
# stdlib imports
import hashlib
import itertools
import json
import logging
import os
import time

from collections import Sequence, Set, OrderedDict, defaultdict
from copy import copy, deepcopy
from gzip import GzipFile
from io import BytesIO
from six import string_types, iteritems

# internal imports
from apt_repoman.connection import Connection
from apt_repoman.repo import KeyExistsError
from apt_repoman import utils

# pypi imports
from botocore.exceptions import ClientError
from pgpy import PGPKeyring
from pydpkg import Dpkg


LOG = logging.getLogger(__name__)


class RepodbError(Exception):
    pass


class InvalidAttributesError(RepodbError):
    pass


class InvalidArchitectureError(RepodbError):
    pass


class InvalidDistributionError(RepodbError):
    pass


class InvalidComponentError(RepodbError):
    pass


class InvalidCopyActionError(RepodbError):
    pass


class ItemExistsError(RepodbError):
    pass


class KeyringNotFoundError(RepodbError):
    pass


class Repodb(object):

    def __init__(self, domain_name, role_arn=None, connection=None):
        self.domain_name = domain_name
        self.role_arn = role_arn
        self._log = LOG or logging.getLogger(__name__)
        self._connection = connection or None
        self._sdb = None
        self._sns = None
        self._meta = {}
        self._domain_exists = None
        self._topic_exists = None
        self._topic_arn = None

    @property
    def connection(self):
        if self._connection is None:
            self._connection = Connection(role_arn=self.role_arn)
        return self._connection

    @property
    def sdb(self):
        if not self._sdb:
            self._sdb = self.connection.sdb
        return self._sdb

    @property
    def sns(self):
        if not self._sns:
            self._sns = self.connection.sns
        return self._sns

    @property
    def meta(self):
        if not self._meta:
            try:
                self._meta = self._get_attributes('meta', always_list=True)
            except KeyError:
                self._log.warning(
                    'No metadata found in simpledb domain %s, '
                    'did you forget to run "repoman setup"?', self.domain_name)
                self._meta = {}
        return self._meta

    @property
    def domain_exists(self):
        if isinstance(self._domain_exists, type(None)):
            domains = []
            paginator = self.sdb.get_paginator('list_domains')
            for page in paginator.paginate():
                domains.extend(page['DomainNames'])
            self._domain_exists = self.domain_name in domains
        return self._domain_exists

    @property
    def topic_arn(self):
        # the SNS API is horrible: there is no way of verifying that a topic
        # exists without iterating through up to 100,000 (!) entries in the
        # output of sns.list_topics(). The recommended means of either
        # verifying an ARN or getting the ARN by the name is...this. :(
        if self._topic_arn is None:
            try:
                if not self.topic_name:
                    return self._topic_arn  # if unset, logging is off
                # be nice and accept either a name or a full ARN here
                elif self.topic_name.startswith('arn:aws:sns:'):
                    self._topic_arn = self.sns.create_topic(
                        Name=self.topic_name.split(':')[5])['TopicArn']
                else:
                    self._topic_arn = self.sns.create_topic(
                        Name=self.topic_name)['TopicArn']
            except KeyError:
                # someone put something unparseable into the topic config
                self._log.warning(
                    'Configured SNS topic "%s" is not parseable as either '
                    'a topic name or a topic ARN; logging is disabled.',
                    self.topic_name)
            except ClientError as ex:
                self._log.warning(
                    'Could not fetch/create SNS topic "%s"; '
                    'logging disabled: %s', self.topic_name, ex)
        return self._topic_arn

    @property
    def archs(self):
        archs = self.meta.get('archs', [])
        # the 'all' architecture is special
        if 'all' not in archs:
            archs.append('all')
        # ditto
        if 'source' not in archs:
            archs.append('source')
        return archs

    @property
    def dists(self):
        return self.meta.get('dists', [])

    @property
    def comps(self):
        return self.meta.get('comps', [])

    @property
    def topic_name(self):
        # this may be unset and that is legit
        try:
            return self.meta.get('topic_name')[0]
        except (IndexError, AttributeError, TypeError):
            return None

    @property
    def origin(self):
        # this may be unset and that is legit
        try:
            return self.meta.get('origin')[0]
        except (IndexError, AttributeError, TypeError):
            return 'repoman'

    @property
    def label(self):
        # this may be unset and that is legit
        try:
            return self.meta.get('label')[0]
        except (IndexError, AttributeError, TypeError):
            return 'repoman'

    def _create_domain(self):
        if self.domain_exists:
            self._log.warning('Simpledb domain "%s" already exists',
                              self.domain_name)
            return None
        try:
            self._log.warning('Creating simpledb domain %s', self.domain_name)
            return self.sdb.create_domain(DomainName=self.domain_name)
        except Exception as ex:
            self._log.fatal('Could not create simpledb domain "%s": %s',
                            self.domain_name, ex)
            raise

    def _create_meta(self, dists=[], comps=[], archs=[],
                     topic_name='', origin='', label='',
                     test_data=''):
        notifications = []
        while 'all' in archs:
            self._log.warning('You cannot add the "all" architecture; like '
                              'the Vorlons it has always been here.')
            archs.remove('all')
        while 'source' in archs:
            self._log.warning('You cannot add the "source" architecture; like '
                              'the Vorlons it has always been here.')
            archs.remove('source')
        for key in ('dists', 'comps', 'archs'):
            targets = locals()[key]
            if key not in self.meta:
                # if not present, add
                self._meta[key] = targets
            else:
                # if present, merge
                if isinstance(self._meta[key], string_types):
                    self._meta[key] = [self._meta[key]]
                existing = set(self._meta[key])
                new = set(targets)
                self._meta[key] = sorted(existing.union(new))
            for target in targets:
                notifications.append(
                    {'action': 'add', 'type': key, 'name': target,
                     'caller': self.connection.caller_id})
        if topic_name:
            self._log.debug('setting up sns notifications: %s', topic_name)
            self._meta['topic_name'] = [topic_name]
            notifications.append(
                {'action': 'add', 'type': 'sns_topic',
                 'name': topic_name, 'caller': self.connection.caller_id})
        if origin:
            self._log.debug('setting repo origin: %s', origin)
            self._meta['origin'] = [origin]
            notifications.append(
                {'action': 'add', 'type': 'origin',
                 'name': origin, 'caller': self.connection.caller_id})
        if label:
            self._log.debug('setting repo label: %s', label)
            self._meta['label'] = [label]
            notifications.append(
                {'action': 'add', 'type': 'label',
                 'name': label, 'caller': self.connection.caller_id})
        if topic_name:
            self._log.debug('setting up sns notifications: %s', topic_name)
            self._meta['topic_name'] = [topic_name]
            notifications.append(
                {'action': 'add', 'type': 'sns_topic',
                 'name': topic_name, 'caller': self.connection.caller_id})
        if test_data:
            self._meta['test_data'] = test_data
        self._log.debug('updated meta: %s', self._meta)
        response = self._put_attributes('meta', self.meta, replace=True)
        if response:
            self._send_notifications(notifications)
        return response

    def _delete_meta(self, dists=[], comps=[], archs=[],
                     topic_name=False, origin=False, label=False,
                     test_data=False):
        notifications = []
        for key in ('archs', 'dists', 'comps'):
            targets = locals()[key]
            for target in targets:
                current = getattr(self, key)
                if target not in current:
                    self._log.warning(
                        'Repo %s do not include %s; cannot delete',
                        key, target)
                    continue
                current.pop(current.index(target))
                self.meta[key] = current
                notifications.append(
                    {'action': 'delete', 'type': key,
                     'name': target, 'caller': self.connection.caller_id})
        if topic_name and 'topic_name' in self.meta:
            self._log.debug('Disabling sns notifications')
            self.meta['topic_name'] = ''
            notifications.append(
                {'action': 'delete', 'type': 'sns_topic',
                 'name': self.topic_name,
                 'caller': self.connection.caller_id})
        if origin and 'origin' in self.meta:
            self._log.debug('Disabling sns notifications')
            self.meta['origin'] = ''
            notifications.append(
                {'action': 'delete', 'type': 'origin',
                 'name': self.origin,
                 'caller': self.connection.caller_id})
        if label and 'label' in self.meta:
            self._log.debug('Disabling sns notifications')
            self.meta['label'] = ''
            notifications.append(
                {'action': 'delete', 'type': 'label',
                 'name': self.label,
                 'caller': self.connection.caller_id})
        if test_data and 'test_data' in self.meta:
            self.meta['test_data'] = ''
        response = self._put_attributes('meta', self.meta, replace=True)
        if response:
            self._send_notifications(notifications)
        return response

    def _send_notifications(self, notifications):
        if not self.topic_arn:
            return None
        for alert in notifications:
            self.sns.publish(
                TopicArn=self.topic_arn,
                Message=json.dumps(alert),
                Subject='Repoman notification')

    def _respool_attributes(self, attributes, replace=False):
        """ Transform a python dictionary in the form
        {str: str} or {str: [str, str]} into a list of
        dicts in the form [{'Name': str, 'Value': str}, ...]

        If the replace argument is not none, update each of the
        response dicts to include 'Replace'=replace.

        :param attributes: list
        :param replace: trinary True/False/None
        :returns list
        """
        response = []
        # simpledb doesn't care about order but our unit tests do :(
        for key in sorted(attributes.keys()):
            val = attributes[key]
            if not isinstance(key, string_types):
                raise InvalidAttributesError(
                    'simpledb attribute names must be strings: %s:%s' %
                    (type(key), key))
            if isinstance(val, string_types):
                entry = {'Name': key, 'Value': val}
                if replace is not None:
                    entry['Replace'] = replace
                response.append(entry)
            # anything that quacks like a list/tuple, we can handle
            elif isinstance(val, (Sequence, Set)):
                for item in val:
                    if isinstance(item, string_types):
                        entry = {'Name': key, 'Value': item}
                        if replace is not None:
                            entry['Replace'] = replace
                        response.append(entry)
                    else:
                        raise InvalidAttributesError(
                            'simpledb attribute values must be strings: %s:%s'
                            % (type(item), item))
            else:
                raise InvalidAttributesError(
                    'I can only convert string or list-like things '
                    'into simpledb attributes: %s:%s', (type(val), val))
        return response

    def _unspool_attributes(self, attributes, always_list=False):
        """ Transform output of boto3.sdb.get_attributes into
            a python dictionary in the form {name: [val, val...]}
        :param attributes: list of {'Name': str, 'Value': str} dicts
        :type attributes: list
        :return: dictionary of key=val[,val].. pairs
        :rtype: dict
        """
        response = {}
        for item in attributes:
            k = item['Name']
            v = item['Value']
            if k in response:
                if type(response[k]) is list:
                    response[k].append(v)
                else:
                    response[k] = [response[k]]
                    response[k].append(v)
            else:
                if always_list:
                    response[k] = [v]
                else:
                    response[k] = v
        return response

    def _get_attributes(self, key, attribute_names=[], consistent_read=True,
                        always_list=False):
        try:
            attributes = self.sdb.get_attributes(
                DomainName=self.domain_name,
                ItemName=key,
                AttributeNames=attribute_names,
                ConsistentRead=consistent_read)['Attributes']
        except Exception as ex:
            self._log.debug(
                'Could not read attributes for item %s: %s', key, ex)
            raise
        return self._unspool_attributes(attributes, always_list)

    def _item_exists(self, key, attribute_names=[]):
        try:
            return self._get_attributes(key, attribute_names)
        except KeyError:
            return None

    def _put_attributes(self, key, attrs, replace=True):
        attributes = self._respool_attributes(attrs, replace)
        try:
            response = self.sdb.put_attributes(
                DomainName=self.domain_name,
                ItemName=key,
                Attributes=attributes)
        except Exception as ex:
            self._log.fatal('Could not update key %s: %s', key, ex)
            raise
        return response

    def _put_item(self, item, replace=True):
        """A convenient wrapper around _put_attributes() and
        _compute_keyname_from_item()"""
        keyname = self._compute_keyname_from_item(item)
        return self._put_attributes(keyname, item, replace)

    def _delete_item(self, item):
        key = self._compute_keyname_from_item(item)
        attrs = self._respool_attributes(item, replace=None)
        try:
            response = self.sdb.delete_attributes(
                DomainName=self.domain_name,
                ItemName=key,
                Attributes=attrs)
        except Exception as ex:
            self._log.fatal('Could not delete item %s: %s', key, ex)
            raise
        return response

    def _select(self, query, consistent_read=True):
        pag = self.sdb.get_paginator('select')
        for page in pag.paginate(
                SelectExpression=query,
                ConsistentRead=consistent_read):
            for item in page.get('Items', []):
                yield self._unspool_attributes(item['Attributes'])

    def _assemble_select_query(self, names=[], dists=[], comps=[], archs=[],
                               versions=[], name_wildcard=False):
        """Query the simpledb database for package items: this function
        assembles a simpledb select query as a string suitable for feeding
        to repodb._select()

        Options are filters: setting e.g. dists to ['jessie', 'xenial'] will
        only return packages in that distribution.  With no options set, this
        will return the entire database.

        :param names: list of package names (strings)
        :param dists: list of repository distributions (strings)
        :param comps: list of repository components (strings)
        :param archs: list of package architectures (strings)
        :param versions: list of package versions (strings)
        :returns: a generator object for a simpledb select() query
        :rtype: Generator
        """
        query = 'select * from `{0}` where `name` is not null'.format(
            self.domain_name)
        selectors = []
        tmpl = "every({0}) in ({1})"
        wild_tmpl = "`{0}` LIKE '{1}%'"
        if names:
            if name_wildcard:
                # Warning: you will verrrry quickly hit the statement
                # predicate limit here; simpledb does not support
                # `every(foo) LIKE ('bar%', 'baz%')`
                statement = ' or '.join(
                    [wild_tmpl.format('name', name) for name in names])
                selectors.append(' {0}'.format(statement))
            else:
                selectors.append(tmpl.format(
                    'name', ','.join(["'%s'" % x for x in names])))
        if dists:
            selectors.append(tmpl.format(
                'distribution', ','.join(["'%s'" % x for x in dists])))
        if comps:
            selectors.append(tmpl.format(
                'component', ','.join(["'%s'" % x for x in comps])))
        if archs:
            selectors.append(tmpl.format(
                'architecture', ','.join(["'%s'" % x for x in archs])))
        if versions:
            selectors.append(tmpl.format(
                'version', ','.join(["'%s'" % x for x in versions])))
        if selectors:
            query += ' and '
            query += ' and '.join(selectors)
        self._log.debug('query: %s', query)
        return query

    def _check_for_hash(self, key):
        if self._get_attributes(key):
            return True
        else:
            return False

    def _compute_keyname(self, name, version, dist, comp, arch):
        """ return a sha256 hex hash of the name+version+dist+comp
        to be used as a unique item key in SDB
        """
        source_str = name + version + dist + comp + arch
        return hashlib.sha256(source_str.encode('ascii')).hexdigest()

    def _compute_keyname_from_item(self, item):
        return self._compute_keyname(
            name=item['name'],
            version=item['version'],
            dist=item['distribution'],
            comp=item['component'],
            arch=item['architecture'])

    def _split_control_text(self, txt, max_elems=256):
        """Simpledb attribute txtues are capped at 1024 bytes, and
        debian control messages can easily be longer than that.
        Even better, control messages can contain unicode strings.

        So: split into a list of substrings, each of which could
        be up to 4 bytes long...
        """
        splits = OrderedDict()
        frags = [txt[i:i+max_elems] for i in range(0, len(txt), max_elems)]
        # pad each key name to the minimum necessary length
        padding = len(str(len(frags)))  # think about it :)
        for count, frag in enumerate(frags):
            splits['controltxt%s' % str(count).zfill(padding)] = frag
        return splits

    def _build_dist_release(self, dist, origin, comps=[], archs=[], date=None):
        self._log.debug('assembling release file for %s', dist)
        if not archs:
            # if you're looking at this and going "wtf?" don't worry,
            # it's not you, it's me. I made a terrible decision in 2011
            # and we all get to pay for it forever
            archs = copy(self.archs)
            archs.remove('all')
        # okay that's done, we can move on now. pretend it never happened.
        if not comps:
            comps = self.comps
        if not date:
            date = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime())
        return (
            'Origin: {origin}\n'
            'Label: {origin}\n'
            'Codename: {dist}\n'
            'Acquire-By-Hash: no\n'
            'Date: {date}\n'
            'Components: {comps}\n'
            'Architectures: {archs}\n'.format(
                origin=origin,
                dist=dist,
                date=date,
                comps=' '.join(comps),
                archs=' '.join(archs),
            ))

    def _create_pkg_msg_from_item(self, item, dist):
        message = 'Filename: pool/%s/%s/%s/%s\n' % (
            dist,
            item['name'][0],
            item['name'],
            item['filename'])
        message += 'MD5sum: %s\n' % item['md5']
        message += 'SHA1: %s\n' % item['sha1']
        message += 'SHA256: %s\n' % item['sha256']
        message += 'Size: %s\n' % item['size']
        # re-assemble control text from all fragments
        # this has to go last, as the control message
        # might have trailing newlines
        control_frags = []
        control_txt = ''
        for attr in item:
            if attr.startswith('controltxt'):
                control_frags.append(attr)
        for attr in sorted(control_frags):
            control_txt += item[attr]
        message += control_txt + '\n'
        return message

    def _build_package_files(self, dists):
        self._log.debug('assembling packages files for %s', dists)
        # the dict always has to return a string at the leaves in
        # order that gzip will have something to work on
        package_files = defaultdict(
            lambda: defaultdict(lambda: defaultdict(lambda: '')))
        # get all packages for the dists we are publishing
        archs = set(self.archs)
        archs.remove('source')
        query = self._create_sorted_package_dict(
            self._select(self._assemble_select_query(
                dists=dists, comps=self.comps, archs=archs)))
        # iterate over every package returned by simpledb;
        # sort them into a nested dictionary:
        # {dist: {comp: {arch: 'Packages.txt'}}}
        # (presort by package name so that output order remains more or less
        # consistent)
        for name in sorted(query.keys()):
            for dist in dists:  # iterate over every dist we are publishing
                for comp in self.comps:  # over every component we know
                    for arch in archs:  # and every architecture
                        if arch == 'all':
                            continue  # these do not get their own section
                        for pkg in query[name][dist][comp][arch]:
                            message = self._create_pkg_msg_from_item(
                                pkg, dist)
                            package_files[dist][comp][arch] += message
                        # packages with architecture=all show up in all
                        # binary distributions
                        for pkg in query[name][dist][comp]['all']:
                            message = self._create_pkg_msg_from_item(
                                pkg, dist)
                            package_files[dist][comp][arch] += message
        return package_files

    def _create_src_msg_from_item(self, item, dist):
        message = 'Directory: pool/%s/%s/%s\n' % (
            dist,
            item['name'][0],
            item['name'])
        message += 'Package: %s\n' % item['name']
        # re-assemble message text from all fragments. this has to go last, as
        # the message might have trailing newlines
        control_frags = []
        control_txt = ''
        for attr in item:
            if attr.startswith('controltxt'):
                control_frags.append(attr)
        for attr in sorted(control_frags):
            control_txt += item[attr]
        message += control_txt + '\n'
        return message

    def _build_source_files(self, dists):
        self._log.debug('assembling sources files for %s', dists)
        # the dict always has to return a string at the leaves in
        # order that gzip will have something to work on
        source_files = defaultdict(
            lambda: defaultdict(lambda: defaultdict(lambda: '')))
        # get all sources for the dists we are publishing
        query = self._create_sorted_package_dict(
            self._select(self._assemble_select_query(
                dists=dists, comps=self.comps, archs=['source'])))
        # iterate over every package returned by simpledb;
        # sort them into a nested dictionary:
        # {dist: {comp: {'source': 'Sources.txt'}}}
        for name in sorted(query.keys()):
            for dist in dists:  # iterate over every dist we are publishing
                for comp in self.comps:  # over every component we know
                    for src in query[name][dist][comp]['source']:
                        message = self._create_src_msg_from_item(src, dist)
                        source_files[dist][comp]['source'] += message
        return source_files

    def _gzip_nested_files(self, package_files):
        """Iterate over a nested dictionary of strings, return
        an equivalently nested dict of gzipped bytes."""
        ret = defaultdict(
            lambda: defaultdict(lambda: defaultdict(lambda: b'')))
        for dist, comps in iteritems(package_files):
            for comp, archs in iteritems(comps):
                for arch in archs:
                    text = package_files[dist][comp][arch]
                    with BytesIO() as buf:
                        with GzipFile(fileobj=buf, mode='wb') as gz:
                            gz.write(text.encode('utf-8'))
                        ret[dist][comp][arch] = buf.getvalue()
        return ret

    def _nested_dict(self, dists=[]):
        """ Many repoman functions return a nested dictionary
        in the form {dist: {comp: {arch: <something>}}}, this
        function provides such a container."""
        ret = {}
        for dist in dists:
            ret[dist] = {}
            for comp in self.comps:
                ret[dist][comp] = {}
                for arch in self.archs:
                    ret[dist][comp][arch] = None
        return ret

    def _generate_dist_release_files(self, dists,
                                     package_files,
                                     package_gz_files,
                                     source_files,
                                     source_gz_files,
                                     origin,
                                     label):
        dist_release_files = dict(itertools.product(dists, [None]))
        for dist in dists:
            dist_release_files[dist] = self._build_dist_release(
                dist, origin)
            for line in self._generate_release_hashes(
                    dist, package_files, package_gz_files,
                    source_files, source_gz_files):
                dist_release_files[dist] += line
        return dist_release_files

    def _generate_leaf_release_files(self, dists, origin, label):
        leaf_release_files = self._nested_dict(dists)
        for dist in dists:
            for comp in self.comps:
                for arch in self.archs:
                    leaf_release_files[dist][comp][arch] = (
                        'Archive: {dist}\n'
                        'Component: {comp}\n'
                        'Origin: {origin}\n'
                        'Label: {label}\n'
                        'Architecture: {arch}\n'.format(
                            dist=dist,
                            comp=comp,
                            origin=origin,
                            label=label,
                            arch=arch))
        return leaf_release_files

    def _get_gpg(self, gpg_home):
        ringfile = os.path.expanduser(
            os.path.join(gpg_home, 'secring.gpg'))
        if not os.path.isfile(ringfile):
            raise KeyringNotFoundError(
                'No gpg secret keyring found at "%s"' % ringfile)
        kr = PGPKeyring()
        kr.load(ringfile)
        return kr

    def _generate_release_sigs(self,
                               gpg_home,
                               gpg_signers,
                               dist_release_files,
                               gpg_passphrases=[]):
        self._log.debug('generating release signatures')
        ret = {}
        gpg = self._get_gpg(gpg_home)
        for dist, text in iteritems(dist_release_files):
            ret[dist] = ''
            for idx, signer in enumerate(gpg_signers):
                with gpg.key(signer) as sec:
                    self._log.debug('signing %s with key %s', dist, signer)
                    if sec.is_unlocked:
                        sig = sec.sign(text, detach=True)
                    else:
                        with sec.unlock(gpg_passphrases[idx]):
                            sig = sec.sign(text, detach=True)
                ret[dist] += str(sig)
        return ret

    def _generate_release_hashes(self, dist, pkgs, pkgs_gz, srcs, srcs_gz):
        for hashname in ('md5', 'sha1', 'sha256'):
            if hashname == 'md5':
                hn = 'MD5Sum'  # grrr
            else:
                hn = hashname.upper()
            yield '{0}:\n'.format(hn)
            hasher = getattr(hashlib, hashname)
            for comp in pkgs[dist]:
                src_text = srcs[dist][comp]['source'].encode('utf-8')  # py27--
                src_gzdata = srcs_gz[dist][comp]['source']
                yield ' {h} {l} {c}/source/Sources\n'.format(
                    h=hasher(src_text).hexdigest(),
                    l=len(src_text),
                    c=comp)
                yield ' {h} {l} {c}/source/Sources.gz\n'.format(
                    h=hasher(src_gzdata).hexdigest(),
                    l=len(src_gzdata),
                    c=comp)
                for arch in pkgs[dist][comp]:
                    pkg_text = pkgs[dist][comp][arch].encode('utf-8')  # py27--
                    pkg_gzdata = pkgs_gz[dist][comp][arch]
                    # add the leaf to the dist-level release file
                    # note the space at the start of this string
                    yield ' {h} {l} {c}/binary-{a}/Packages\n'.format(
                        h=hasher(pkg_text).hexdigest(),
                        l=len(pkg_text),
                        c=comp,
                        a=arch)
                    yield ' {h} {l} {c}/binary-{a}/Packages.gz\n'.format(
                        h=hasher(pkg_gzdata).hexdigest(),
                        l=len(pkg_gzdata),
                        c=comp,
                        a=arch)
        yield '\n'

    def _assemble_path_data(self, dist_release_files, dist_release_sigs,
                            package_files, package_gz_files,
                            source_files, source_gz_files,
                            leaf_release_files):
        path_data = []
        suffix = (self._connection.profile_name, self._connection.role_arn)
        # assemble lists of paths to write to s3
        for dist in dist_release_files.keys():
            # dist_release_files and dist_release_sigs are only
            # keyed by dist name
            path_data.append(
                ('dists/{0}/Release'.format(dist),
                 dist_release_files[dist]) + suffix)
            if dist_release_sigs:
                path_data.append(('dists/{0}/Release.gpg'.format(dist),
                                  dist_release_sigs[dist]) + suffix)
            # everything else is a walk down the comps/archs tree
            for comp in self.comps:
                for arch in self.archs:
                    if arch == 'all':
                        continue  # we don't generate a specific release here
                    elif arch == 'source':
                        path_data.append(
                            ('dists/{0}/{1}/source/Sources'.format(dist, comp),
                             source_files[dist][comp][arch]) + suffix)
                        path_data.append(
                            ('dists/{0}/{1}/source/Sources.gz'.format(
                                dist, comp),
                             source_gz_files[dist][comp][arch]) + suffix)
                        path_data.append(
                            ('dists/{0}/{1}/source/Release'.format(
                                dist, comp),
                             leaf_release_files[dist][comp][arch]) + suffix)
                    else:
                        path_data.append(
                            ('dists/{0}/{1}/binary-{2}/Packages'.format(
                                dist, comp, arch),
                             package_files[dist][comp][arch]) + suffix)
                        path_data.append(
                            ('dists/{0}/{1}/binary-{2}/Packages.gz'.format(
                                dist, comp, arch),
                             package_gz_files[dist][comp][arch]) + suffix)
                        path_data.append(
                            ('dists/{0}/{1}/binary-{2}/Release'.format(
                                dist, comp, arch),
                             leaf_release_files[dist][comp][arch]) + suffix)
        return path_data

    def _create_sorted_package_dict(self, sources, latest_versions=0):
        """ Given a list of package items returned from query(), sort
        them into a nested dict in the form:
            {name: {dist: {comp: {arch: [item, item...]}}}}
        ...then sort each list of items at the tree leaves from
        lowest to highest by debian package version sorting order."""
        self._log.debug('latest: %d', latest_versions)
        sorted_sources = defaultdict(lambda: defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))))
        for pkg in sources:
            sorted_sources[pkg['name']][
                pkg['distribution']][
                    pkg['component']][pkg['architecture']].append(pkg)
        for name, dists in iteritems(sorted_sources):
            for dist, comps in iteritems(dists):
                for comp, archs in iteritems(comps):
                    for arch, pkgs in iteritems(archs):
                        ct = len(pkgs)
                        pkgs.sort(key=lambda x:
                                  Dpkg.compare_versions_key(x['version']))
                        self._log.debug('len pkgs: %d', ct)
                        if latest_versions > 0:
                            # prune to only the N most "recent" packages
                            sorted_sources[name][dist][comp][arch] = (
                                pkgs[-latest_versions:])
                        elif (latest_versions < 0 and -latest_versions-1 < ct):
                            # prune to ALL BUT the N most "recent" packages
                            sorted_sources[name][dist][comp][arch] = (
                                pkgs[:latest_versions])
        return sorted_sources

    def _check_spec(self, sources, targets):
        """
        Ensure that two nested dicts of package items have the same keys and
        the same number of items in each list.
        """
        success = True
        for name, dists in iteritems(sources):
            if name not in targets:
                self._log.fatal('Source package name %s not found in '
                                'destination spec: %s', name, targets)
                success = False
                break
            for dist, comps in iteritems(dists):
                for comp, archs in iteritems(comps):
                    for arch, items in iteritems(archs):
                        if len(items) != len(targets[name][dist][comp][arch]):
                            self._log.fatal(
                                'Source and destination specs for '
                                'package %s have non-matching lengths. '
                                '\n\tSrc: %s\n\tDst: %s ',
                                name, items, targets[name][dist][comp][arch])
                            success = False
                            break
        if not success:
            raise InvalidCopyActionError(
                'Something has gone horrifically wrong; the '
                'source and destination side of a copy action '
                'should have the same number of items. This is '
                'probably a bug: please open an issue on github '
                'with as much detail as you can provide')
        return True

    def _walk_ndcai(self, ndcai, enumerate_items=False):
        """Multiple functions in this object have to unspool
        a nested dictionary in the form of:
            {name: {dist: {comp: {arch: [item, item]}}}}
        This leads to some ridiculously deep nesting of functions,
        so turn this process into a generator."""
        for name, dists in iteritems(ndcai):
            for dist, comps in iteritems(dists):
                for comp, archs in iteritems(comps):
                    for arch, items in iteritems(archs):
                        if enumerate_items:
                            for idx, item in enumerate(items):
                                yield (name, dist, comp, arch, idx, item)
                        else:
                            for item in items:
                                yield (name, dist, comp, arch, item)

    def initialize(self, dists=[], comps=[], archs=[], topic_name=None,
                   origin=None, label=None):
        self._log.info('Creating simpledb domain')
        self._create_domain()
        self._log.info('Initializing repository database')
        self._create_meta(dists=dists, comps=comps, archs=archs,
                          topic_name=topic_name,
                          origin=origin,
                          label=label)
        return 0

    def add_meta(self, archs=[], dists=[], comps=[],
                 topic_name='', origin='', label='',
                 test_data=''):
        if archs:
            self._log.info('Adding architectures: %s', archs)
        if dists:
            self._log.info('Adding distributions: %s', dists)
        if comps:
            self._log.info('Adding components: %s', comps)
        if topic_name:
            self._log.info('Adding SNS topic for logging: %s', topic_name)
        if origin:
            self._log.info('Setting repository origin: %s', origin)
        if label:
            self._log.info('Setting repository label: %s', label)
        if test_data:
            self._log.info('Writing test data to repo: %s', test_data)
        return self._create_meta(dists, comps, archs,
                                 topic_name, origin, label, test_data)

    def rm_meta(self, archs=[], dists=[], comps=[],
                topic_name=False, test_data=False):
        if archs:
            self._log.info('Deleting architectures: %s', archs)
        if dists:
            self._log.info('Deleting distributions: %s', dists)
        if comps:
            self._log.info('Deleting components: %s', comps)
        if topic_name:
            self._log.info('Deleting SNS topic for logging')
        if test_data:
            self._log.info('Deleting test data from repo')
        return self._delete_meta(dists, comps, archs, topic_name, test_data)

    def find_invalid_metadata(self, candidates, metadata_type):
        """Check that each member of a list of possible repository
           metadata configuration values (architectures, distributions,
           components) are known to the repository."""
        if not getattr(self, metadata_type):
            raise KeyError(
                '%s is not a known type of metadata' % metadata_type)
        failed = []
        for candidate in candidates:
            if candidate not in getattr(self, metadata_type):
                self._log.debug(
                    '%s is not a known type of %s', candidate, metadata_type)
                failed.append(candidate)
        return failed

    def check_valid_archs(self, archs=[]):
        """Assert that each member of the list `archs` is a architecture
        this repo is currently configured to serve; otherwise raise
        InvalidArchitectureError.

        :param archs: list
        :returns: true
        :raises: InvalidArchitectureError
        """
        unrecognized = self.find_invalid_metadata(archs, 'archs')
        if unrecognized:
            raise InvalidArchitectureError(
                'architectures %s are not currently served by this repo' %
                unrecognized)
        return True

    def check_valid_dists(self, dists=[]):
        """Assert that each member of the list `dists` is a distribution
        this repo is currently configured to serve; otherwise raise
        InvalidDistributionError.

        :param dists: list
        :returns: true
        :raises: InvalidDistributionError
        """
        unrecognized = self.find_invalid_metadata(dists, 'dists')
        if unrecognized:
            raise InvalidDistributionError(
                'distributions %s are not currently served by this repo' %
                unrecognized)
        return True

    def check_valid_comps(self, comps=[]):
        """Assert that each member of the list `comps` is a component
        this repo is currently configured to serve; otherwise raise
        InvalidComponentError.

        :param comps: list
        :returns: true
        :raises: InvalidComponentError
        """
        unrecognized = self.find_invalid_metadata(comps, 'comps')
        if unrecognized:
            raise InvalidComponentError(
                'components %s are not currently served by this repo' %
                unrecognized)
        return True

    def add_package(self, pkg, dists=[], comps=[],
                    overwrite=False, auto_purge=0):
        """Import the metadata from a pydpkg.Dpkg object as a
        simpledb item for each of the specified distributions
        and components.

        :param pkg: a pydpkg.Dpkg object
        :param dists: list of strings
        :param comps: list of strings
        :param overwrite: bool
        :param auto_purge: int
        """
        pkg_name = pkg.get_header('package')
        pkg_file = os.path.basename(pkg.filename)
        control_str = str(pkg.message.as_string())
        pkg_arch = pkg.get_header('architecture')
        self.check_valid_archs([pkg_arch])
        self.check_valid_dists(dists)
        self.check_valid_comps(comps)
        for dist in dists:
            for comp in comps:
                attrs = {'name': pkg_name,
                         'filename': pkg_file,
                         'distribution': dist,
                         'component': comp,
                         'version': pkg.version,
                         'architecture': pkg_arch,
                         'md5': pkg.md5,
                         'sha1': pkg.sha1,
                         'sha256': pkg.sha256,
                         'size': str(pkg.filesize)}
                attrs.update(self._split_control_text(control_str))
                self._log.debug('attrs: %s', attrs)
                key_name = self._compute_keyname_from_item(attrs)
                self._log.debug('key name: %s', key_name)
                if self._item_exists(key_name) and not overwrite:
                    raise ItemExistsError(
                        'Package %s version %s in distribution % and '
                        'component %s already exists in simpledb',
                        pkg_name, pkg.version, dist, comp)
                self._put_attributes(key_name, attrs)
                self._send_notifications([
                    {'action': 'add', 'type': 'package',
                     'name': attrs['name'],
                     'version': attrs['version'],
                     'distribution': attrs['distribution'],
                     'component': attrs['component'],
                     'caller': self.connection.caller_id}])
                if auto_purge > 0:
                    self._log.warning(
                        'Automatically purging %d oldest versions of %s '
                        'in the %s distribution, %s component and %s '
                        'architecture.', auto_purge, pkg_name, dist,
                        comp, pkg_arch)
                    targets = self.get_candidates(
                        dist, comp, names=[pkg_name], archs=[pkg_arch],
                        latest_versions=-auto_purge)
                    self.do_rm(targets)

    def add_source(self, dsc, dists=[], comps=[], overwrite=False,
                   auto_purge=0):
        """Import the metadata from a pydpkg.Dsc object as a
        simpledb item for each of the specified distributions
        and components.

        Annoyingly, debian dsc messages are juuuuust different
        enough from binary ones that this has to be a special case.

        :param dsc: a pydpkg.Dsc object
        :param dists: list of strings
        :param comps: list of strings
        :param overwrite: bool
        :param auto_purge: int
        """
        # blow up immediately if the source bundle is incomplete
        # or corrupt
        dsc.validate()
        dsc_arch = 'source'
        dsc_name = dsc.source
        source_files = [os.path.basename(x) for x in dsc.source_files]
        message_str = dsc.message_str
        self.check_valid_dists(dists)
        self.check_valid_comps(comps)
        for dist in dists:
            for comp in comps:
                attrs = {'name': dsc_name,
                         'files': source_files,  # nb: this will be a list
                         'distribution': dist,
                         'component': comp,
                         'version': dsc.version,
                         'architecture': dsc_arch}
                attrs.update(self._split_control_text(message_str))
                self._log.debug('attrs: %s', attrs)
                key_name = self._compute_keyname_from_item(attrs)
                self._log.debug('key name: %s', key_name)
                if self._item_exists(key_name) and not overwrite:
                    raise ItemExistsError(
                        'Source package %s version %s in distribution % and '
                        'component %s already exists in simpledb',
                        dsc_name, dsc.version, dist, comp)
                self._put_attributes(key_name, attrs)
                self._send_notifications([
                    {'action': 'add', 'type': 'source',
                     'name': dsc_name,
                     'version': attrs['version'],
                     'distribution': dist,
                     'component': comp,
                     'caller': self.connection.caller_id}])
                if auto_purge > 0:
                    self._log.warning(
                        'Automatically purging %d oldest versions of %s '
                        'in the %s distribution, %s component and %s '
                        'architecture.', auto_purge, dsc_name, dist,
                        comp, dsc_arch)
                    targets = self.get_candidates(
                        dist, comp, names=[dsc_name], archs=[dsc_arch],
                        latest_versions=-auto_purge)
                    self.do_rm(targets)

    def publish(self, repo, dists=[],
                gpg_home='~/.gnupg', gpg_signers=[], gpg_passphrases=[]):
        """Assemble the metadata files of the repository and write them
        to the repo s3 bucket."""
        retval = 0
        origin = self.origin or 'repoman'
        label = self.label or 'repoman'
        if dists is None or len(dists) == 0:
            dists = self.dists
        package_files = self._build_package_files(dists)
        source_files = self._build_source_files(dists)
        # pre-compress the package file strings
        package_gz_files = self._gzip_nested_files(package_files)
        source_gz_files = self._gzip_nested_files(source_files)
        dist_release_files = self._generate_dist_release_files(
            dists, package_files, package_gz_files,
            source_files, source_gz_files,
            origin, label)
        leaf_release_files = self._generate_leaf_release_files(
            dists, origin, label)
        if gpg_signers:
            dist_release_sigs = self._generate_release_sigs(
                gpg_home, gpg_signers, dist_release_files, gpg_passphrases)
        else:
            dist_release_sigs = None

        path_data = self._assemble_path_data(
            dist_release_files, dist_release_sigs,
            package_files, package_gz_files,
            source_files, source_gz_files,
            leaf_release_files)

        results = utils.write_paths(
            repo.bucket_name, path_data, threads=0)

        for path, code in results:
            if not code or code.get(
                    'ResponseMetadata', {}).get('HTTPStatusCode') != 200:
                self._log.error(
                    'Did not successfully write "s3://%s/%s: %s',
                    repo.bucket_name, path, code)
                retval = 1

        self._log.info('Successfully published repository for dists %s '
                       'to bucket s3://%s', dists, repo.bucket_name)

        return retval

    def query(self, names=[], dists=[], comps=[], archs=[], versions=[],
              latest_versions=0, name_wildcard=False):
        """
        A friendly wrapper around repodb._assemble_select_query() and
        _create_sorted_package_dict() that returns a nested dictionary of
        package items, keyed by package name, and with the list of packages
        sorted in strict debian package ordering, oldest to newest:
            {name:{dist:{comp:[pkg, pkg]}}}
        Optionally if latest_versions is positive, each of the lists is pruned
        to only the N newest packages (again by debian sorting rules).

        :param names: list of strings
        :param dists: list of strings
        :param comps: list of strings
        :param archs: list of strings
        :param versions: list of strings
        :param latest_versions: int
        :param name_wildcard: bool
        :rtype: dict
        """
        if dists:
            self.check_valid_dists(dists)
            if isinstance(dists, str):
                dists = [dists]
        if comps:
            self.check_valid_comps(comps)
            if isinstance(comps, str):
                comps = [comps]
        if archs:
            self.check_valid_archs(archs)
            if isinstance(archs, str):
                archs = [archs]
        query_str = self._assemble_select_query(
            names=names,
            dists=dists,
            comps=comps,
            archs=archs,
            versions=versions,
            name_wildcard=name_wildcard)
        return self._create_sorted_package_dict(
            self._select(query_str),
            latest_versions)

    def get_candidates(self, src_dist, src_comp,
                       names=[], versions=[], archs=[],
                       latest_versions=0, name_wildcard=False):
        """
        A small wrapper around repodb.query() that ensures we are only
        passing in a single distribution and/or component, since
        multiples would not make sense in a copying context.

        :param src_dict: string
        :param src_comp: string
        :param packages: list of strings
        :param versions: list of strings
        :param archs: list of strings
        :param latest_versions: int
        :param name_wildcard: bool
        :rtype: dict
        :raises: InvalidDistributionError, InvalidComponentError
        """
        if not isinstance(src_dist, string_types):
            raise InvalidDistributionError(
                'src_dist must be the name of a single distribution: %s',
                src_dist)
        if not isinstance(src_comp, string_types):
            raise InvalidComponentError(
                'src_comp must be the name of a single component: %s',
                src_comp)
        sources = self.query(
            names=names,
            dists=[src_dist] if src_dist else [],
            comps=[src_comp] if src_comp else [],
            archs=archs,
            versions=versions,
            latest_versions=latest_versions,
            name_wildcard=name_wildcard)
        return sources

    def get_copy_spec(self, candidates, src_dist, src_comp,
                      dst_dist=None, dst_comp=None,
                      prune_for_promote=False):
        """Given a nested dict of source packages for copying,
        compute an equivalent nested dict of new package items to
        create, and prune no-ops from both the source and the
        destination side.

        If prune_for_promote is true, prune from the source side
        any packages that are older, version-wise, than the newest
        package on the destination side.
        """
        # if no destination distribution or component is specified, then
        # the move is within the source dist/comp
        dst_dist = dst_dist or src_dist
        dst_comp = dst_comp or src_comp
        self.check_valid_dists([dst_dist])
        self.check_valid_comps([dst_comp])
        existing = self.get_candidates(
            dst_dist, dst_comp, names=candidates.keys())
        targets = defaultdict(
            lambda: defaultdict(
                lambda: defaultdict(lambda: defaultdict(list))))
        # what's an o(n^4) between friends?
        for name, dist, comp, arch, idx, old in self._walk_ndcai(
                candidates, enumerate_items=True):
            new = deepcopy(old)
            new['distribution'] = dst_dist or new['distribution']
            new['component'] = dst_comp or new['component']
            if new == old:
                self._log.debug('Same as source: %s', new)
                candidates[name][dist][comp][arch][idx] = None
                continue
            if new in existing[name][dist][comp][arch]:
                self._log.debug('Already at target: %s', new)
                candidates[name][dist][comp][arch][idx] = None
                continue
            if prune_for_promote:
                # if the old version < the newest available
                # version on the destination side (if there are
                # any), we are not a candidate for promotion
                if existing[name][dst_dist][dst_comp][arch]:
                    latest = existing[name][dst_dist][dst_comp][arch][-1]
                    res = Dpkg.compare_versions(
                        latest['version'], new['version'])
                    if res >= 0:
                        self._log.warning('skipping')
                        candidates[name][dist][comp][arch][idx] = None
                        continue
            targets[name][dist][comp][arch].append(new)
            # prune no-ops from source side
            if None in candidates[name][dist][comp][arch]:
                candidates[name][dist][comp][arch].remove(None)
        if self._check_spec(candidates, targets):
            return candidates, targets

    def do_copy(self, candidates, targets, repo,
                overwrite=False, auto_purge=0):
        for name, dist, comp, arch, idx, pkg in self._walk_ndcai(
                targets, enumerate_items=True):
            src_dist = candidates[name][dist][comp][arch][idx]['distribution']
            src_comp = candidates[name][dist][comp][arch][idx]['component']
            dst_dist = pkg['distribution']
            if src_dist != dst_dist:
                self._log.warning(
                    'copy of package %s from distribution %s to '
                    '%s requires an s3 copy',
                    name, src_dist, dst_dist)
                # handle both source and binary files
                filenames = filter(
                    None,
                    itertools.chain([pkg.get('filename')],
                                    pkg.get('files', [])))
                for fn in filenames:
                    old_path = os.path.join(
                        'pool', src_dist, pkg['name'][0],
                        pkg['name'], fn)
                    new_path = os.path.join(
                        'pool', dst_dist, pkg['name'][0],
                        pkg['name'], fn)
                    try:
                        repo.copy_key(old_path, new_path, overwrite)
                    except KeyExistsError:
                        self._log.warning(
                            'package already exists at s3://%s ',
                            new_path)
            self._log.info(
                'creating package %s version %s '
                'distribution %s component %s '
                'architecture %s', pkg['name'],
                pkg['version'], pkg['distribution'],
                pkg['component'], pkg['architecture'])
            key = self._compute_keyname_from_item(pkg)
            self._put_attributes(key, pkg)
            self._send_notifications([
                {'action': 'copy', 'type': 'package',
                 'name': pkg['name'],
                 'version': pkg['version'],
                 'dst_distribution': pkg['distribution'],
                 'dst_component': pkg['component'],
                 'src_distribution': src_dist,
                 'src_component': src_comp,
                 'caller': self.connection.caller_id}])
        if auto_purge > 0:
            for name, dists in iteritems(targets):
                for dist, comps in iteritems(dists):
                    for comp, archs in iteritems(comps):
                        for arch, items in iteritems(archs):
                            # horrible cheat here
                            dst_dist = items[0]['distribution']
                            dst_comp = items[0]['component']
                            self._log.warning(
                                'Automatically purging %d oldest versions of '
                                '%s in the %s distribution, %s component and '
                                '%s architecture.', auto_purge, name, dist,
                                comp, arch)
                            purge_targets = self.get_candidates(
                                dst_dist, dst_comp, names=[name], archs=[arch],
                                latest_versions=-auto_purge)
                            self.do_rm(purge_targets)

    def do_rm(self, targets):
        for name, dist, comp, arch, item in self._walk_ndcai(targets):
            self._log.warning(
                'Deleting pkg %s version %s in distribution '
                '%s component %s architecture %s',
                item['name'], item['version'],
                item['distribution'], item['component'],
                item['architecture'])
            self._delete_item(item)
            self._send_notifications([
                {'action': 'delete', 'type': 'package',
                 'name': item['name'],
                 'version': item['version'],
                 'distribution': item['distribution'],
                 'component': item['component'],
                 'caller': self.connection.caller_id}])




# stdlib imports
import logging
import time
from multiprocessing import Pool

from apt_repoman.connection import Connection

LOG = logging.getLogger(__name__)


def write_path(bucket_name, path, data, profile_name='', role_arn=''):
    now = time.time()
    connection = Connection(
        profile_name=profile_name,
        role_arn=role_arn)
    s3 = connection.session.resource('s3')
    LOG.info('writing s3://%s/%s', bucket_name, path)
    if path.endswith('gz'):
        content_type = 'binary/octet-stream'
    else:
        content_type = 'text/plain'

    k = s3.Object(bucket_name, path)
    elapsed = time.time() - now
    LOG.debug('wrote %s/%s in %f sec', bucket_name, path, elapsed)
    result = k.put(Body=data, ContentType=content_type)
    return path, result


def func_star(args):
    """Convert `f([1,2])` to `f(1,2)`"""
    return write_path(*args)


def write_paths(bucket_name, tups, threads=0):
    if threads == 0:
        threads = len(tups)
    LOG.debug('threads: %s', threads)
    pool = Pool()
    args = [(bucket_name,) + tup for tup in tups]
    now = time.time()
    LOG.debug('dispatching writers')
    results = pool.map(func_star, args)
    LOG.debug('all dispatched in %s seconds', time.time() - now)
    now = time.time()
    LOG.debug('closing writer pool')
    pool.close()
    LOG.debug('closed in %s seconds', time.time() - now)
    now = time.time()
    LOG.debug('joining writer pool')
    pool.join()
    LOG.debug('joined in %s seconds', time.time() - now)
    LOG.info('done writing to s3')
    return results



#!/usr/bin/env python

import json
import time
import unittest
import os

from collections import OrderedDict
from gzip import GzipFile
from io import BytesIO
from mock import patch, PropertyMock, MagicMock

import botocore.session
from botocore.stub import Stubber, ANY

from apt_repoman.repodb import Repodb
from apt_repoman.repodb import InvalidAttributesError
from apt_repoman.repodb import InvalidCopyActionError

HASH = 'ad30985578dcf4e5fe0d8f40270fcff7b4e39720307f95b4511be0eda8ddc0b9'

IPSUM = """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasellus
mollis hendrerit quam, non consectetur elit vestibulum sed. Donec pharetra
egestas purus eu venenatis. Etiam dignissim pretium metus. Suspendisse nec dui
at nisi consectetur feugiat molestie eu metus. Donec eget urna id lorem ornare
aliquet eget et felis. Integer rutrum, eros ac vehicula aliquam, lorem quam
tincidunt mauris, eget laoreet diam nisi a libero. In elementum dui faucibus
odio efficitur, a dignissim ante pulvinar. Donec auctor mi nunc, commodo
hendrerit nulla consequat et. Nulla facilisi. Pellentesque tempor dui at
ultrices facilisis. In ullamcorper at lacus non luctus. Integer faucibus,
ligula in venenatis blandit, leo neque hendrerit velit, ac porttitor urna metus
a dolor."""


class RepodbTest(unittest.TestCase):

    def setUp(self):
        self.repodb = Repodb('testdomain')
        self._dir = os.path.dirname(__file__)

    def testMeta(self):
        self.repodb._meta = {'dists': ['d1', 'd2'],
                             'comps': ['c1', 'c2'],
                             'archs': ['a1', 'a2']}
        self.assertEqual(self.repodb.dists, ['d1', 'd2'])
        self.assertEqual(self.repodb.comps, ['c1', 'c2'])
        # tricky...
        self.assertEqual(self.repodb.archs, ['a1', 'a2', 'all', 'source'])

    def testRespoolAttributes(self):
        _in_good = {'foo': 'bar', 'xyzzy': ['bada', 'bing']}
        _in_bad_key = {1: 'bar'}
        _in_bad_val = {'a': 1}
        _in_bad_subval = {'a': ['bar', 1]}
        _out_true = [
            {'Name': 'foo', 'Value': 'bar', 'Replace': True},
            {'Name': 'xyzzy', 'Value': 'bada', 'Replace': True},
            {'Name': 'xyzzy', 'Value': 'bing', 'Replace': True}]
        _out_false = [
            {'Name': 'foo', 'Value': 'bar', 'Replace': False},
            {'Name': 'xyzzy', 'Value': 'bada', 'Replace': False},
            {'Name': 'xyzzy', 'Value': 'bing', 'Replace': False}]
        _out_none = [
            {'Name': 'foo', 'Value': 'bar'},
            {'Name': 'xyzzy', 'Value': 'bada'},
            {'Name': 'xyzzy', 'Value': 'bing'}]
        self.assertEqual(
            self.repodb._respool_attributes(_in_good, True),
            _out_true)
        self.assertEqual(
            self.repodb._respool_attributes(_in_good, False),
            _out_false)
        self.assertEqual(
            self.repodb._respool_attributes(_in_good, None),
            _out_none)
        self.assertRaises(
            InvalidAttributesError,
            self.repodb._respool_attributes, _in_bad_key)
        self.assertRaises(
            InvalidAttributesError,
            self.repodb._respool_attributes, _in_bad_val)
        self.assertRaises(
            InvalidAttributesError,
            self.repodb._respool_attributes, _in_bad_subval)

    def testUnspoolAttributes(self):
        _in = [{'Name': 'foo', 'Value': 'bar'},
               {'Name': 'xyzzy', 'Value': 'bada'},
               {'Name': 'xyzzy', 'Value': 'bing'}]
        _out = {'foo': 'bar', 'xyzzy': ['bada', 'bing']}
        self.assertEqual(
            self.repodb._unspool_attributes(_in),
            _out)

    def testAssembleSelectQuery(self):
        self.assertEqual(
            self.repodb._assemble_select_query(),
            "select * from `testdomain` where `name` is not null")
        self.assertEqual(
            self.repodb._assemble_select_query(names=['foo']),
            "select * from `testdomain` where `name` is not null and "
            "every(name) in ('foo')")
        self.assertEqual(
            self.repodb._assemble_select_query(names=['foo', 'bar']),
            "select * from `testdomain` where `name` is not null and "
            "every(name) in ('foo','bar')")
        self.assertEqual(
            self.repodb._assemble_select_query(
                names=['foo'], name_wildcard=True),
            "select * from `testdomain` where `name` is not null and  "
            "`name` LIKE 'foo%'")
        self.assertEqual(
            self.repodb._assemble_select_query(
                names=['foo', 'bar'], name_wildcard=True),
            "select * from `testdomain` where `name` is not null and  "
            "`name` LIKE 'foo%' or `name` LIKE 'bar%'")
        self.assertEqual(
            self.repodb._assemble_select_query(names=['foo'],
                                               dists=['bar', 'baz']),
            "select * from `testdomain` where `name` is not null and "
            "every(name) in ('foo') and every(distribution) in ('bar','baz')")
        self.assertEqual(
            self.repodb._assemble_select_query(names=['foo'],
                                               comps=['bar', 'baz']),
            "select * from `testdomain` where `name` is not null and "
            "every(name) in ('foo') and every(component) in ('bar','baz')")
        self.assertEqual(
            self.repodb._assemble_select_query(names=['foo'],
                                               archs=['bar', 'baz']),
            "select * from `testdomain` where `name` is not null and "
            "every(name) in ('foo') and every(architecture) in ('bar','baz')")
        self.assertEqual(
            self.repodb._assemble_select_query(names=['foo'],
                                               versions=['bar', 'baz']),
            "select * from `testdomain` where `name` is not null and "
            "every(name) in ('foo') and every(version) in ('bar','baz')")

    def testComputeKeyname(self):
        self.assertEqual(
            self.repodb._compute_keyname('foo', 'bar', 'baz', 'qux', 'xyzzy'),
            HASH)

    def testComputKeynameFromItem(self):
        _in = {'name': 'foo', 'version': 'bar',
               'distribution': 'baz', 'component': 'qux',
               'architecture': 'xyzzy'}
        self.assertEqual(
            self.repodb._compute_keyname_from_item(_in),
            HASH)

    def testSplitControlText(self):
        self.assertEqual(
            self.repodb._split_control_text(IPSUM, 64),
            OrderedDict([
                ('controltxt00', 'Lorem ipsum dolor sit amet, consectetur a'
                                 'dipiscing elit. Phasell'),
                ('controltxt01', 'us\nmollis hendrerit quam, non consectetu'
                                 'r elit vestibulum sed. D'),
                ('controltxt02', 'onec pharetra\negestas purus eu venenatis'
                                 '. Etiam dignissim pretiu'),
                ('controltxt03', 'm metus. Suspendisse nec dui\nat nisi con'
                                 'sectetur feugiat molesti'),
                ('controltxt04', 'e eu metus. Donec eget urna id lorem orna'
                                 're\naliquet eget et feli'),
                ('controltxt05', 's. Integer rutrum, eros ac vehicula aliqu'
                                 'am, lorem quam\ntincidun'),
                ('controltxt06', 't mauris, eget laoreet diam nisi a libero'
                                 '. In elementum dui fauc'),
                ('controltxt07', 'ibus\nodio efficitur, a dignissim ante pu'
                                 'lvinar. Donec auctor mi '),
                ('controltxt08', 'nunc, commodo\nhendrerit nulla consequat '
                                 'et. Nulla facilisi. Pell'),
                ('controltxt09', 'entesque tempor dui at\nultrices facilisi'
                                 's. In ullamcorper at lac'),
                ('controltxt10', 'us non luctus. Integer faucibus,\nligula '
                                 'in venenatis blandit, le'),
                ('controltxt11', 'o neque hendrerit velit, ac porttitor urn'
                                 'a metus\na dolor.')])
        )

    def testBuildDistRelease(self):
        date = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime())
        self.assertEqual(
            self.repodb._build_dist_release(
                'foo', 'bar',
                comps=['baz', 'qux'],
                archs=['bada', 'bing'],
                date=date),
            '\n'.join(['Origin: bar',
                       'Label: bar',
                       'Codename: foo',
                       'Acquire-By-Hash: no',
                       'Date: %s' % date,
                       'Components: baz qux',
                       'Architectures: bada bing\n']))

    def testCreatePackageMessageFromItem(self):
        _in = {
            'name': 'foo',
            'filename': 'bar',
            'md5': 'DEADBEEF',
            'sha1': 'BEEFCAFE',
            'sha256': 'CAFEFACE',
            'size': '123',
            'controltxt01': 'us\nmollis hendrerit quam, non consectetur elit vestibulum sed. D',
            'controltxt00': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasell',
            'controltxt02': 'onec pharetra\negestas purus eu venenatis. Etiam dignissim pretiu'}
        self.assertEqual(
            self.repodb._create_pkg_msg_from_item(_in, 'xyzzy'),
            '\n'.join([
                'Filename: pool/xyzzy/f/foo/bar',
                'MD5sum: DEADBEEF',
                'SHA1: BEEFCAFE',
                'SHA256: CAFEFACE',
                'Size: 123',
                'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasellus',
                'mollis hendrerit quam, non consectetur elit vestibulum sed. Donec pharetra',
                'egestas purus eu venenatis. Etiam dignissim pretiu\n']))

    def testCreateSourceMessageFromItem(self):
        _in = {
            'name': 'foo',
            'controltxt01': 'us\nmollis hendrerit quam, non consectetur elit vestibulum sed. D',
            'controltxt00': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasell',
            'controltxt02': 'onec pharetra\negestas purus eu venenatis. Etiam dignissim pretiu'}
        expected = '\n'.join([
                'Directory: pool/xyzzy/f/foo',
                'Package: foo',
                'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasellus',
                'mollis hendrerit quam, non consectetur elit vestibulum sed. Donec pharetra',
                'egestas purus eu venenatis. Etiam dignissim pretiu\n'])
        self.assertEqual(
            expected,
            self.repodb._create_src_msg_from_item(_in, 'xyzzy'))

    def testGzipPackageFiles(self):
        # fun!
        _in = {'trusty': {'main': {'amd64': 'foobar'}},
               'xenial': {'main': {'all': 'xyzzy'}}}
        _out = self.repodb._gzip_nested_files(_in)
        tma = _out['trusty']['main']['amd64']
        xma = _out['xenial']['main']['all']
        with BytesIO(tma) as bz:
            with GzipFile(fileobj=bz) as gz:
                self.assertEqual(b'foobar', gz.read())
        with BytesIO(xma) as bz:
            with GzipFile(fileobj=bz) as gz:
                self.assertEqual(b'xyzzy', gz.read())

    @patch('apt_repoman.repodb.Repodb.archs', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.comps', new_callable=PropertyMock)
    def testNestedDict(self, comps, archs):
        comps.return_value = ['c1', 'c2']
        archs.return_value = ['a1', 'a2']
        _out = {'d1': {'c1': {'a1': None, 'a2': None},
                       'c2': {'a1': None, 'a2': None}},
                'd2': {'c1': {'a1': None, 'a2': None},
                       'c2': {'a1': None, 'a2': None}}}
        _in = self.repodb._nested_dict(['d1', 'd2'])
        self.assertEqual(_in, _out)

    @patch('apt_repoman.repodb.Repodb.archs', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.comps', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.dists', new_callable=PropertyMock)
    def testGenerateDistReleaseFiles(self, dists, comps, archs):
        dists.return_value = ['d1']
        comps.return_value = ['c1']
        archs.return_value = ['a1', 'all']
        faketime = time.gmtime(1497895073.870057)
        with patch('time.gmtime', return_value=faketime):
            package_files = {'d1': {'c1': {'a1': 'foo'}}}
            package_gz_files = {'d1': {'c1': {'a1': b'0xDEADBEEF'}}}
            source_files = {'d1': {'c1': {'source': 'bar'}}}
            source_gz_files = {'d1': {'c1': {'source': b'0xBEEFCAFE'}}}
            origin = 'test'
            label = 'test'
            self.maxDiff = None
            expected = {'d1': """Origin: test
Label: test
Codename: d1
Acquire-By-Hash: no
Date: Mon, 19 Jun 2017 17:57:53 +0000
Components: c1
Architectures: a1
MD5Sum:
 37b51d194a7513e45b56f6524f2d51f2 3 c1/source/Sources
 a4d4f03fbcc4a36782648488dd07319f 10 c1/source/Sources.gz
 acbd18db4cc2f85cedef654fccc4a4d8 3 c1/binary-a1/Packages
 545882e2eba6b126518d07c954698c83 10 c1/binary-a1/Packages.gz
SHA1:
 62cdb7020ff920e5aa642c3d4066950dd1f01f4d 3 c1/source/Sources
 cef0d0350de7697c4abe1e6b7db788d46dc748b1 10 c1/source/Sources.gz
 0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33 3 c1/binary-a1/Packages
 fbc02cbc52f9aa96fefa06c567b180df6df832db 10 c1/binary-a1/Packages.gz
SHA256:
 fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9 3 c1/source/Sources
 09a597d7489048ac580a36b0381b35d9140738f8e3b8dfc19a5edfd41f09cc2b 10 c1/source/Sources.gz
 2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae 3 c1/binary-a1/Packages
 5fce5f5878415a3bae17db60a69b08d51f0f962228bfdbb462993a3ac16171e4 10 c1/binary-a1/Packages.gz

"""}
            returned = self.repodb._generate_dist_release_files(
                ['d1'],
                package_files, package_gz_files,
                source_files, source_gz_files,
                origin, label)
            self.assertEqual(expected, returned)

    @patch('apt_repoman.repodb.Repodb.archs', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.comps', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.dists', new_callable=PropertyMock)
    def testGenerateLeafReleaseFiles(self, dists, comps, archs):
        dists.return_value = ['d1']
        comps.return_value = ['c1']
        archs.return_value = ['a1', 'all']
        origin = 'test'
        label = 'test'
        self.maxDiff = None
        self.assertEqual(
            self.repodb._generate_leaf_release_files(['d1'], origin, label),
            {'d1': {'c1': {'a1': 'Archive: d1\nComponent: c1\nOrigin: test\nLabel: test\nArchitecture: a1\n',
             'all': 'Archive: d1\nComponent: c1\nOrigin: test\nLabel: test\nArchitecture: all\n'}}}
        )

    @patch('apt_repoman.repodb.Repodb.archs', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.comps', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.dists', new_callable=PropertyMock)
    def testBuildPackageFiles(self, dists, comps, archs):
        dists.return_value = ['xenial', 'jessie']
        comps.return_value = ['main', 'nightly']
        archs.return_value = ['source', 'i386', 'amd64', 'all']
        self.repodb._sdb = botocore.session.get_session().create_client('sdb', region_name='us-east-1')
        self.maxDiff = None
        with open(os.path.join(self._dir, 'query_output.json')) as fp:
            select_response = json.loads(fp.read())
        with Stubber(self.repodb._sdb) as stub:
            stub.add_response('select', select_response, {
                'ConsistentRead': True,
                'SelectExpression': ANY})
            expected = {
                "jessie": {
                  "main": {
                    "i386": "Filename: pool/jessie/t/testdeb/3:testdeb_1.1.0-test2_all.deb\nMD5sum: 85813d0688494314fa37b4bca6939782\nSHA1: efa19aee680fbbb98f1915ee14c18ef902dfa025\nSHA256: 4d3ef2fe026ee7fae63584e2361ec3176f24fe448f48925189bb60fb2e72f2c6\nSize: 1298\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\n",
                    "amd64": "Filename: pool/jessie/t/testdeb/3:testdeb_1.1.0-test2_amd64.deb\nMD5sum: 568728506dd4374cb34c12dbf9e9a148\nSHA1: dedbe9b395b22ff18d161bb9caaabaea09b2da5b\nSHA256: 4ab8523efef0a343f8cf301d67d4cb71b76b7761f1023523940928c223d02437\nSize: 1296\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: amd64\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/jessie/t/testdeb/3:testdeb_1.1.0-test2_all.deb\nMD5sum: 85813d0688494314fa37b4bca6939782\nSHA1: efa19aee680fbbb98f1915ee14c18ef902dfa025\nSHA256: 4d3ef2fe026ee7fae63584e2361ec3176f24fe448f48925189bb60fb2e72f2c6\nSize: 1298\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\n"
                  }
                },
                "xenial": {
                  "nightly": {
                    "i386": "Filename: pool/xenial/t/testdeb/testdeb_1:0.0.0-test_all.deb\nMD5sum: 149e61536a9fe36374732ec95cf7945d\nSHA1: a5d28ae2f23e726a797349d7dd5f21baf8aa02b4\nSHA256: 547500652257bac6f6bc83f0667d0d66c8abd1382c776c4de84b89d0f550ab7f\nSize: 910\nPackage: testdeb\nVersion: 1:0.0.0-test\nSection: base\nPriority: extra\nArchitecture: all\nInstalled-Size: 0\nMaintainer: Nathan Mehl <n@climate.com>\nDescription: testdeb\n a bogus debian package for testing dpkg builds\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.0.0-test_all.deb\nMD5sum: c9a1e0a35cc3706e8003edbc8bf60e7f\nSHA1: e3dc33ad3aa9c3ff81020bdfd629e171f6ac2c0e\nSHA256: 6bfe58c48ab34585b8c1bc671e8103c499e17af2a3a4e237a58bd57fc8d1501d\nSize: 1304\nPackage: testdeb\nVersion: 1:1.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test_all.deb\nMD5sum: 9d72f9a5e0d739dcc5241f7e50e5addd\nSHA1: a923bd5489b8e5ec7d3e96fb1b700c29d81725df\nSHA256: a8e6bca9936270a057a670658bd4e43aaf1791495734d3d666be12815b4a56e2\nSize: 1300\nPackage: testdeb\nVersion: 1:1.1.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test2_all.deb\nMD5sum: 36611a3c94464e954e54e5c385ff647e\nSHA1: f58ebb9e618d3ce1293c1a6e7dcb7e1d30ca242f\nSHA256: 7323ab214df42f89a5df4eedc4c0aeece14bff31425f2dd7bd90d2d21d56163a\nSize: 1292\nPackage: testdeb\nVersion: 1:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/2:testdeb_0.0.0-test_all.deb\nMD5sum: c8cd10216d5e99a18971c80531d10b01\nSHA1: 2e3f6aca44459898fd1e05ab659ba1cd0aa9a613\nSHA256: 7f29b9631765fb459c046fbda1586767e4d8a9ad330dff86bfae91c684aa3f30\nSize: 1294\nPackage: testdeb\nVersion: 2:0.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/3:testdeb_1.1.0-test2_all.deb\nMD5sum: 85813d0688494314fa37b4bca6939782\nSHA1: efa19aee680fbbb98f1915ee14c18ef902dfa025\nSHA256: 4d3ef2fe026ee7fae63584e2361ec3176f24fe448f48925189bb60fb2e72f2c6\nSize: 1298\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\n",
                    "amd64": "Filename: pool/xenial/t/testdeb/3:testdeb_1.1.0-test2_amd64.deb\nMD5sum: 568728506dd4374cb34c12dbf9e9a148\nSHA1: dedbe9b395b22ff18d161bb9caaabaea09b2da5b\nSHA256: 4ab8523efef0a343f8cf301d67d4cb71b76b7761f1023523940928c223d02437\nSize: 1296\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: amd64\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/testdeb_1:0.0.0-test_all.deb\nMD5sum: 149e61536a9fe36374732ec95cf7945d\nSHA1: a5d28ae2f23e726a797349d7dd5f21baf8aa02b4\nSHA256: 547500652257bac6f6bc83f0667d0d66c8abd1382c776c4de84b89d0f550ab7f\nSize: 910\nPackage: testdeb\nVersion: 1:0.0.0-test\nSection: base\nPriority: extra\nArchitecture: all\nInstalled-Size: 0\nMaintainer: Nathan Mehl <n@climate.com>\nDescription: testdeb\n a bogus debian package for testing dpkg builds\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.0.0-test_all.deb\nMD5sum: c9a1e0a35cc3706e8003edbc8bf60e7f\nSHA1: e3dc33ad3aa9c3ff81020bdfd629e171f6ac2c0e\nSHA256: 6bfe58c48ab34585b8c1bc671e8103c499e17af2a3a4e237a58bd57fc8d1501d\nSize: 1304\nPackage: testdeb\nVersion: 1:1.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test_all.deb\nMD5sum: 9d72f9a5e0d739dcc5241f7e50e5addd\nSHA1: a923bd5489b8e5ec7d3e96fb1b700c29d81725df\nSHA256: a8e6bca9936270a057a670658bd4e43aaf1791495734d3d666be12815b4a56e2\nSize: 1300\nPackage: testdeb\nVersion: 1:1.1.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test2_all.deb\nMD5sum: 36611a3c94464e954e54e5c385ff647e\nSHA1: f58ebb9e618d3ce1293c1a6e7dcb7e1d30ca242f\nSHA256: 7323ab214df42f89a5df4eedc4c0aeece14bff31425f2dd7bd90d2d21d56163a\nSize: 1292\nPackage: testdeb\nVersion: 1:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/2:testdeb_0.0.0-test_all.deb\nMD5sum: c8cd10216d5e99a18971c80531d10b01\nSHA1: 2e3f6aca44459898fd1e05ab659ba1cd0aa9a613\nSHA256: 7f29b9631765fb459c046fbda1586767e4d8a9ad330dff86bfae91c684aa3f30\nSize: 1294\nPackage: testdeb\nVersion: 2:0.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/3:testdeb_1.1.0-test2_all.deb\nMD5sum: 85813d0688494314fa37b4bca6939782\nSHA1: efa19aee680fbbb98f1915ee14c18ef902dfa025\nSHA256: 4d3ef2fe026ee7fae63584e2361ec3176f24fe448f48925189bb60fb2e72f2c6\nSize: 1298\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\n"
                  },
                  "main": {
                    "i386": "Filename: pool/xenial/h/hashicorp-consul-template/hashicorp-consul-template_0.15-tcc01_i386.deb\nMD5sum: 8fce38500395227d83095c2e1ef1f473\nSHA1: 88e9312553b083de908c875a2d581d6a407f00b0\nSHA256: baaa65ff64c1b3d1d90f48f87cedf49112a209d8b3af80b8b9422059c42caf00\nSize: 1342\nPackage: hashicorp-consul-template\nVersion: 0.15-tcc01\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: i386\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/testdeb_1:0.0.0-test_all.deb\nMD5sum: 149e61536a9fe36374732ec95cf7945d\nSHA1: a5d28ae2f23e726a797349d7dd5f21baf8aa02b4\nSHA256: 547500652257bac6f6bc83f0667d0d66c8abd1382c776c4de84b89d0f550ab7f\nSize: 910\nPackage: testdeb\nVersion: 1:0.0.0-test\nSection: base\nPriority: extra\nArchitecture: all\nInstalled-Size: 0\nMaintainer: Nathan Mehl <n@climate.com>\nDescription: testdeb\n a bogus debian package for testing dpkg builds\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.0.0-test_all.deb\nMD5sum: c9a1e0a35cc3706e8003edbc8bf60e7f\nSHA1: e3dc33ad3aa9c3ff81020bdfd629e171f6ac2c0e\nSHA256: 6bfe58c48ab34585b8c1bc671e8103c499e17af2a3a4e237a58bd57fc8d1501d\nSize: 1304\nPackage: testdeb\nVersion: 1:1.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test_all.deb\nMD5sum: 9d72f9a5e0d739dcc5241f7e50e5addd\nSHA1: a923bd5489b8e5ec7d3e96fb1b700c29d81725df\nSHA256: a8e6bca9936270a057a670658bd4e43aaf1791495734d3d666be12815b4a56e2\nSize: 1300\nPackage: testdeb\nVersion: 1:1.1.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test2_all.deb\nMD5sum: 36611a3c94464e954e54e5c385ff647e\nSHA1: f58ebb9e618d3ce1293c1a6e7dcb7e1d30ca242f\nSHA256: 7323ab214df42f89a5df4eedc4c0aeece14bff31425f2dd7bd90d2d21d56163a\nSize: 1292\nPackage: testdeb\nVersion: 1:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/2:testdeb_0.0.0-test_all.deb\nMD5sum: c8cd10216d5e99a18971c80531d10b01\nSHA1: 2e3f6aca44459898fd1e05ab659ba1cd0aa9a613\nSHA256: 7f29b9631765fb459c046fbda1586767e4d8a9ad330dff86bfae91c684aa3f30\nSize: 1294\nPackage: testdeb\nVersion: 2:0.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/3:testdeb_1.1.0-test2_all.deb\nMD5sum: 85813d0688494314fa37b4bca6939782\nSHA1: efa19aee680fbbb98f1915ee14c18ef902dfa025\nSHA256: 4d3ef2fe026ee7fae63584e2361ec3176f24fe448f48925189bb60fb2e72f2c6\nSize: 1298\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\n",
                    "amd64": "Filename: pool/xenial/h/hashicorp-consul-template/hashicorp-consul-template_0.15-tcc01_amd64.deb\nMD5sum: 3fe3996aea835bb0fd9c9a7595ee969e\nSHA1: 85b006304079b32cd10283f55d009ba51a79b480\nSHA256: 77a3eaee90d8edabf0a73b3d3e7148d298e4e0461b3a1737b23dc78fce6da22e\nSize: 3521222\nPackage: hashicorp-consul-template\nVersion: 0.15-tcc01\nLicense: unknown\nVendor: none\nArchitecture: amd64\nMaintainer: <n@salacious.local>\nInstalled-Size: 12194\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/3:testdeb_1.1.0-test2_amd64.deb\nMD5sum: 568728506dd4374cb34c12dbf9e9a148\nSHA1: dedbe9b395b22ff18d161bb9caaabaea09b2da5b\nSHA256: 4ab8523efef0a343f8cf301d67d4cb71b76b7761f1023523940928c223d02437\nSize: 1296\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: amd64\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/testdeb_1:0.0.0-test_all.deb\nMD5sum: 149e61536a9fe36374732ec95cf7945d\nSHA1: a5d28ae2f23e726a797349d7dd5f21baf8aa02b4\nSHA256: 547500652257bac6f6bc83f0667d0d66c8abd1382c776c4de84b89d0f550ab7f\nSize: 910\nPackage: testdeb\nVersion: 1:0.0.0-test\nSection: base\nPriority: extra\nArchitecture: all\nInstalled-Size: 0\nMaintainer: Nathan Mehl <n@climate.com>\nDescription: testdeb\n a bogus debian package for testing dpkg builds\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.0.0-test_all.deb\nMD5sum: c9a1e0a35cc3706e8003edbc8bf60e7f\nSHA1: e3dc33ad3aa9c3ff81020bdfd629e171f6ac2c0e\nSHA256: 6bfe58c48ab34585b8c1bc671e8103c499e17af2a3a4e237a58bd57fc8d1501d\nSize: 1304\nPackage: testdeb\nVersion: 1:1.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test_all.deb\nMD5sum: 9d72f9a5e0d739dcc5241f7e50e5addd\nSHA1: a923bd5489b8e5ec7d3e96fb1b700c29d81725df\nSHA256: a8e6bca9936270a057a670658bd4e43aaf1791495734d3d666be12815b4a56e2\nSize: 1300\nPackage: testdeb\nVersion: 1:1.1.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/1:testdeb_1.1.0-test2_all.deb\nMD5sum: 36611a3c94464e954e54e5c385ff647e\nSHA1: f58ebb9e618d3ce1293c1a6e7dcb7e1d30ca242f\nSHA256: 7323ab214df42f89a5df4eedc4c0aeece14bff31425f2dd7bd90d2d21d56163a\nSize: 1292\nPackage: testdeb\nVersion: 1:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/2:testdeb_0.0.0-test_all.deb\nMD5sum: c8cd10216d5e99a18971c80531d10b01\nSHA1: 2e3f6aca44459898fd1e05ab659ba1cd0aa9a613\nSHA256: 7f29b9631765fb459c046fbda1586767e4d8a9ad330dff86bfae91c684aa3f30\nSize: 1294\nPackage: testdeb\nVersion: 2:0.0.0-test\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\nFilename: pool/xenial/t/testdeb/3:testdeb_1.1.0-test2_all.deb\nMD5sum: 85813d0688494314fa37b4bca6939782\nSHA1: efa19aee680fbbb98f1915ee14c18ef902dfa025\nSHA256: 4d3ef2fe026ee7fae63584e2361ec3176f24fe448f48925189bb60fb2e72f2c6\nSize: 1298\nPackage: testdeb\nVersion: 3:1.1.0-test2\nLicense: unknown\nVendor: n@C02SV1F1G8WL\nArchitecture: all\nMaintainer: <n@C02SV1F1G8WL>\nInstalled-Size: 0\nSection: default\nPriority: extra\nHomepage: http://example.com/no-uri-given\nDescription: no description given\n\n\n"
                  }
                }
              }
            returned = self.repodb._build_package_files(dists=['xenial', 'jessie'])
            self.assertEqual(expected, returned)

    @patch('apt_repoman.repodb.Repodb.archs', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.comps', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.dists', new_callable=PropertyMock)
    def testBuildSourceFiles(self, dists, comps, archs):
        dists.return_value = ['xenial', 'jessie']
        comps.return_value = ['main', 'nightly']
        archs.return_value = ['source', 'i386', 'amd64', 'all']
        self.repodb._sdb = botocore.session.get_session().create_client('sdb', region_name='us-east-1')
        self.maxDiff = None
        with open(os.path.join(self._dir, 'query_output.json')) as fp:
            select_response = json.loads(fp.read())
        with Stubber(self.repodb._sdb) as stub:
            stub.add_response('select', select_response, {
                'ConsistentRead': True,
                'SelectExpression': ANY})
            expected = {
                "jessie": {
                  "main": {
                    "source": "Directory: pool/jessie/a/apt-transport-s3\nPackage: apt-transport-s3\nFormat: 3.0 (quilt)\nSource: apt-transport-s3\nBinary: apt-transport-s3\nArchitecture: all\nVersion: 1.2.1-1\nMaintainer: Marcin Kulisz (kuLa) <debian@kulisz.net>\nUploaders: David Watson <dwatson@debian.org>\nHomepage: https://github.com/BashtonLtd/apt-transport-s3\nStandards-Version: 3.9.6\nVcs-Browser: https://github.com/BashtonLtd/apt-transport-s3/tree/debian/sid\nVcs-Git: https://github.com/BashtonLtd/apt-transport-s3.git\nBuild-Depends: python (>= 2.6.6-3), debhelper (>= 9)\nPackage-List: apt-transport-s3 deb admin optional arch=all\nChecksums-Sha1: fcaab2d352dda3c6e8894afef7ecf62f05367637 53764\n apt-transport-s3_1.2.1.orig.tar.gz\n 5833e6929583f1038c6aa2a26d264ed49f09106d 41920\n apt-transport-s3_1.2.1-1.debian.tar.xz\n 36c3bf0921f13de8176b3a86ecff1cac5532ff73 1986 apt-transport-s3_1.2.1-1.dsc\nChecksums-Sha256: 3935a698d3ca56ff02ffe133578c67d1e22b6599ab72643ac3b2564f7786dfcf 53764\n apt-transport-s3_1.2.1.orig.tar.gz\n 3b95f5222410637602f01d3120b0047a187afb8605e6fc444675aa599abc04c1 41920\n apt-transport-s3_1.2.1-1.debian.tar.xz\n fa801cc1da5a9acbbebdce30a1f32b260b50452125d84eedd45b917504206584 1986\n apt-transport-s3_1.2.1-1.dsc\nFiles: 8dfb207307b5fcfc917d5fbc6003ff95 53764\n apt-transport-s3_1.2.1.orig.tar.gz\n 76eb07b79e5cf34ba966a5cdd2817086 41920 apt-transport-s3_1.2.1-1.debian.tar.xz\n eca184b019fc93b069ae4bbe94ab9e3b 1986 apt-transport-s3_1.2.1-1.dsc\n\n\n"
                  }
                },
                "xenial": {
                  "main": {
                    "source": "Directory: pool/xenial/a/apt-transport-s3\nPackage: apt-transport-s3\nFormat: 3.0 (quilt)\nSource: apt-transport-s3\nBinary: apt-transport-s3\nArchitecture: all\nVersion: 1.2.1-1\nMaintainer: Marcin Kulisz (kuLa) <debian@kulisz.net>\nUploaders: David Watson <dwatson@debian.org>\nHomepage: https://github.com/BashtonLtd/apt-transport-s3\nStandards-Version: 3.9.6\nVcs-Browser: https://github.com/BashtonLtd/apt-transport-s3/tree/debian/sid\nVcs-Git: https://github.com/BashtonLtd/apt-transport-s3.git\nBuild-Depends: python (>= 2.6.6-3), debhelper (>= 9)\nPackage-List: apt-transport-s3 deb admin optional arch=all\nChecksums-Sha1: fcaab2d352dda3c6e8894afef7ecf62f05367637 53764\n apt-transport-s3_1.2.1.orig.tar.gz\n 5833e6929583f1038c6aa2a26d264ed49f09106d 41920\n apt-transport-s3_1.2.1-1.debian.tar.xz\n 36c3bf0921f13de8176b3a86ecff1cac5532ff73 1986 apt-transport-s3_1.2.1-1.dsc\nChecksums-Sha256: 3935a698d3ca56ff02ffe133578c67d1e22b6599ab72643ac3b2564f7786dfcf 53764\n apt-transport-s3_1.2.1.orig.tar.gz\n 3b95f5222410637602f01d3120b0047a187afb8605e6fc444675aa599abc04c1 41920\n apt-transport-s3_1.2.1-1.debian.tar.xz\n fa801cc1da5a9acbbebdce30a1f32b260b50452125d84eedd45b917504206584 1986\n apt-transport-s3_1.2.1-1.dsc\nFiles: 8dfb207307b5fcfc917d5fbc6003ff95 53764\n apt-transport-s3_1.2.1.orig.tar.gz\n 76eb07b79e5cf34ba966a5cdd2817086 41920 apt-transport-s3_1.2.1-1.debian.tar.xz\n eca184b019fc93b069ae4bbe94ab9e3b 1986 apt-transport-s3_1.2.1-1.dsc\n\n\n"
                  }
                }
            }
            returned = self.repodb._build_source_files(dists=['xenial', 'jessie'])
            self.assertEqual(expected, returned)

    @patch('apt_repoman.repodb.Repodb.archs', new_callable=PropertyMock)
    @patch('apt_repoman.repodb.Repodb.comps', new_callable=PropertyMock)
    def testAssemblePathData(self, comps, archs):
        self.repodb._connection = MagicMock
        self.repodb._connection.profile_name = 'profile'
        self.repodb._connection.role_arn = 'role'
        comps.return_value = ['c1']
        archs.return_value = ['a1', 'all', 'source']
        dist_release_files = {'d1': 'foo'}
        dist_release_sigs = {'d1': {'c1': {'a1': '--PGP--', 'all': '--PGP--', 'source': '--PGP--'}}}
        package_files = {'d1': {'c1': {'a1': 'packages', 'all': 'morepackages'}}}
        package_gz_files = {'d1': {'c1': {'a1': b'0xDEADBEEF', 'all': b'0xBEEFCAFE'}}}
        source_files = {'d1': {'c1': {'source': 'sources'}}}
        source_gz_files = {'d1': {'c1': {'source': b'0xDEADBEEF'}}}
        leaf_release_files = {'d1': {'c1': {'a1': 'wash', 'all': 'wind', 'source': 'watch'}}}
        expected = [('dists/d1/Release', 'foo', 'profile', 'role'),
                    ('dists/d1/Release.gpg', {'c1': {'a1': '--PGP--', 'all': '--PGP--', 'source': '--PGP--'}}, 'profile', 'role'),
                    ('dists/d1/c1/binary-a1/Packages', 'packages', 'profile', 'role'),
                    ('dists/d1/c1/binary-a1/Packages.gz', b'0xDEADBEEF', 'profile', 'role'),
                    ('dists/d1/c1/binary-a1/Release', 'wash', 'profile', 'role'),
                    ('dists/d1/c1/source/Sources', 'sources', 'profile', 'role'),
                    ('dists/d1/c1/source/Sources.gz', b'0xDEADBEEF', 'profile', 'role'),
                    ('dists/d1/c1/source/Release', 'watch', 'profile', 'role')]
        returned = self.repodb._assemble_path_data(
            dist_release_files, dist_release_sigs,
            package_files, package_gz_files,
            source_files, source_gz_files,
            leaf_release_files)
        self.maxDiff = None
        self.assertEquals(expected, returned)

    def testCreateSortedPackageDict(self):
        _in = [
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-test1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-~est1'},
        ]
        _out = {'foo': {'d1': {'c1': {'a1': [
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-~est1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-test1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'}]}}}}
        _twolatest = {'foo': {'d1': {'c1': {'a1': [
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-test1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'}]}}}}
        _onelatest = {'foo': {'d1': {'c1': {'a1': [
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'}]}}}}
        self.assertEqual(
            json.dumps(self.repodb._create_sorted_package_dict(_in)),
            json.dumps(_out))
        self.assertEqual(
            json.dumps(self.repodb._create_sorted_package_dict(_in, 1)),
            json.dumps(_onelatest))
        self.assertEqual(
            json.dumps(self.repodb._create_sorted_package_dict(_in, 2)),
            json.dumps(_twolatest))

    def testCheckSpec(self):
        _left = {'foo': {'d1': {'c1': {'a1': [
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-test1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'}]}}}}
        _right = {'foo': {'d1': {'c1': {'a1': [
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-test1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'}]}}}}
        _badlist = {'foo': {'d1': {'c1': {'a1': [
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'}]}}}}
        _badkeys = {'bar': {'d1': {'c1': {'a1': [
            {'name': 'bar',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '0.0.0-test1'},
            {'name': 'foo',
             'distribution': 'd1',
             'component': 'c1',
             'architecture': 'a1',
             'version': '1:0.0.0-test1'}]}}}}
        self.maxDiff = None
        self.assertEqual(self.repodb._check_spec(_left, _right), True)
        self.assertRaises(InvalidCopyActionError,
                          self.repodb._check_spec, _left, _badkeys)
        self.assertRaises(InvalidCopyActionError,
                          self.repodb._check_spec, _left, _badlist)

    def testWalkNdcai(self):
        _in = {'foo': {'d1': {'c1': {'a1': ['foo-1', 'foo-2']}}}}
        _out = [('foo', 'd1', 'c1', 'a1', 'foo-1'),
                ('foo', 'd1', 'c1', 'a1', 'foo-2')]
        _out_idx = [('foo', 'd1', 'c1', 'a1', 0, 'foo-1'),
                    ('foo', 'd1', 'c1', 'a1', 1, 'foo-2')]
        self.assertEqual(
            list(self.repodb._walk_ndcai(_in)),
            _out)
        self.assertEqual(
            list(self.repodb._walk_ndcai(_in, True)),
            _out_idx)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(RepodbTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python

from __future__ import print_function

# stdlib imports
import json
import logging
import os
import sys
import time

from logging import config
from pkg_resources import resource_stream
from six import iteritems
from six.moves import input

# pypi imports
from colors import color
from pgpy.errors import PGPDecryptionError
from pydpkg import Dpkg
from pydpkg import Dsc
from pysectools.pinentry import Pinentry
from tabulate import tabulate

# internal imports
from apt_repoman.config import Config
from apt_repoman.connection import Connection
from apt_repoman.repo import KeyExistsError
from apt_repoman.repo import Repo
from apt_repoman.repodb import InvalidArchitectureError
from apt_repoman.repodb import ItemExistsError
from apt_repoman.repodb import Repodb


LOG = logging.getLogger(__name__)
HEADERS = ['name', 'distribution', 'component', 'architecture', 'version']


def repo_print_config(repodb, repo):
    LOG.info('Current repo configuration:')
    LOG.info('\tSimpledb domain: %s', repodb.domain_name)
    LOG.info('\tS3 bucket: s3://%s', repo.bucket_name)
    topic = repodb.topic_name or '---None configured---'
    LOG.info('\tSNS Notification topic: %s', topic)
    LOG.info('\tDistributions: %s', repodb.dists)
    LOG.info('\tComponents: %s', repodb.comps)
    LOG.info('\tArchitectures: %s', repodb.archs)
    LOG.info('\tOrigin: %s', repodb.origin)
    LOG.info('\tLabel: %s', repodb.label)
    return 0


def repo_add(thing, args, repodb, repo):
    abbrev = thing[0:4] + 's'
    argname = thing + '_names'
    candidates = getattr(args, argname)
    existing = getattr(repodb, abbrev)
    # the "all" architecture is special, you cannot remove it
    if abbrev == 'archs':
        while 'all' in args.architecture_names:
            LOG.error('You cannot add the "all" architecture, like the '
                      'Vorlons it has always been here.')
            args.architecture_names.remove('all')
        while 'source' in args.architecture_names:
            LOG.error('You cannot add the "source" architecture, like the '
                      'Vorlons it has always been here.')
            args.architecture_names.remove('source')
    for name in candidates:
        try:
            if name in existing:
                LOG.warning('%s %s already exists in repo %s',
                            thing, name, args.simpledb_domain)
                continue
        except KeyError:
            LOG.fatal('No repo configuration found, did you forget to '
                      'run "repoman setup"?')
            sys.exit(1)
    kwargs = {abbrev: candidates}
    LOG.warning(color(
        'Adding %s(s): %s', fg='red'), thing, ','.join(candidates))
    if confirm(args):
        res = repodb.add_meta(**kwargs)
    LOG.debug(res)
    return repo_print_config(repodb, repo)


def repo_rm(thing, args, repodb, repo):
    abbrev = thing[0:4] + 's'
    argname = thing + '_names'
    candidates = getattr(args, argname)
    existing = getattr(repodb, abbrev)
    elected = []
    evil = []
    if abbrev == 'archs':
        while 'all' in args.architecture_names:
            LOG.error('You cannot remove the "all" architecture, like the '
                      'Talamasca it watches and is always there.')
            args.architecture_names.remove('all')
        while 'source' in args.architecture_names:
            LOG.error('You cannot remove the "source" architecture, like the '
                      'Talamasca it watches and is always there.')
            args.architecture_names.remove('source')
    for name in candidates:
        try:
            if name not in existing:
                LOG.warning('%s %s does not exist in repo %s',
                            thing, name, args.simpledb_domain)
                continue
            elected.append(name)
        except KeyError:
            LOG.fatal('No repo configuration found, did you forget to '
                      'run "repoman setup"?')
            sys.exit(1)
    if not elected:
        LOG.error('No valid %ss to delete', thing)
        return 1
    LOG.warning(color(
        'Deleting %s(s): %s', fg='red'), thing, ','.join(elected))
    # ugh I am so sorry
    evil_msg = 'You are about to delete EVERY %s this repo serves' % thing
    if abbrev == 'archs':
        if len(elected) == len(repodb.archs) - 1:
            evil.append(evil_msg)
    elif len(elected) == len(existing):
        evil.append(evil_msg)
    if confirm(args, evil):
        res = repodb.rm_meta(**{abbrev: elected})
    LOG.debug(res)
    LOG.warning('Deleting a {0} from a repository does not actually '
                'delete the package files from S3 or the items from '
                'simpledb, it just stops including those {0}s in '
                'query results, cp/rm commands and '
                'the publishing process: you will need to clean up '
                'the orphaned resources manually if you care.'.format(
                    thing))
    return repo_print_config(repodb, repo)


def repo_topic(action, args, repodb, repo):
    if action == 'add':
        topic_name = args.topic_name[0]
        LOG.warning(color('Setting up SNS topic "%s" for logging' %
                          topic_name, fg='red'))
        if confirm(args):
            repodb.add_meta(topic_name=topic_name)
    elif action == 'rm':
        LOG.warning(color('Disabling SNS logging!', fg='red'))
        if confirm(args):
            repodb.rm_meta(topic_name=True)
    return repo_print_config(repodb, repo)


def repo_origin(action, args, repodb, repo):
    if action == 'add':
        origin = args.origin[0]
        LOG.warning(
            color('Setting repository origin: "%s"' % origin, fg='red'))
        if confirm(args):
            repodb.add_meta(origin=origin)
    else:
        LOG.fatal('You cannot remove an origin, you can only change it.')
        sys.exit(1)
    return repo_print_config(repodb, repo)


def repo_label(action, args, repodb, repo):
    if action == 'add':
        label = args.label[0]
        LOG.warning(
            color('Setting repository label: "%s"' % label, fg='red'))
        if confirm(args):
            repodb.add_meta(label=label)
    else:
        LOG.fatal('You cannot remove an label, you can only change it.')
        sys.exit(1)
    return repo_print_config(repodb, repo)


def repo(args, repodb, repo):
    command = args.repo_command
    if command == 'show-config':
        return repo_print_config(repodb, repo)
    if command.startswith('add-') or command.startswith('rm-'):
        verb, noun = command.split('-')
        if noun in ('topic', 'origin', 'label'):
            # sns topic handling is a little different
            return globals()['repo_' + noun](verb, args, repodb, repo)
        else:
            return globals()['repo_' + verb](noun, args, repodb, repo)


def add(args, repodb, repo):
    """Add packages"""
    # we check this here before we risk uploading to s3
    if not validate_meta(args, repodb):
        return 1
    success = 0
    for fn in args.files:
        LOG.info('attempting to add file: %s', fn)
        if fn.endswith('.deb'):
            pkg = Dpkg(fn)
        elif fn.endswith('.dsc'):
            pkg = Dsc(fn)
        else:
            LOG.error('File "%s" is neither a deb nor a dsc files', fn)
            success += 1
            continue
        try:
            if isinstance(pkg, Dsc):
                LOG.info('attempting to add source to s3: %s',
                         os.path.basename(fn))
                repo.add_source(pkg,
                                dists=args.distribution,
                                overwrite=args.overwrite)
                LOG.info('attempting to add source to simpledb: %s',
                         os.path.basename(fn))
                repodb.add_source(pkg,
                                  dists=args.distribution,
                                  comps=args.component,
                                  overwrite=args.overwrite,
                                  auto_purge=args.auto_purge)
            else:
                arch = pkg.architecture
                repodb.check_valid_archs([arch])
                LOG.info('attempting to add package to s3: %s',
                         os.path.basename(pkg.filename))
                repo.add_package(pkg,
                                 dists=args.distribution,
                                 overwrite=args.overwrite)
                LOG.info('attempting to add package to simpledb: %s',
                         os.path.basename(pkg.filename))
                repodb.add_package(pkg,
                                   dists=args.distribution,
                                   comps=args.component,
                                   overwrite=args.overwrite,
                                   auto_purge=args.auto_purge)
            LOG.info('Successfully added %s to repoman!', fn)
        except InvalidArchitectureError:
            LOG.error('Package %s is built for the "%s" architecture, '
                      'which this repo is not currently configured to '
                      'serve; I will not add it.  You may which to run '
                      '"repoman repo add_architecture %s"', fn, arch, arch)
            success += 1
            continue
        except KeyExistsError:
            LOG.error('Package %s already exists in S3, you either want the '
                      '--overwrite flag or you want to move/copy the package '
                      'within the repo.  Skipping.', fn)
            success += 1
            continue
        except ItemExistsError:
            LOG.error('Package %s already exists in simpledb, you either want '
                      'the --overwrite flag or you want to move/copy the '
                      'package within the repo.  Skipping.', fn)
            success += 1
            continue
    if success > 0:
        LOG.error('Not all packages uploadeded successfully; inspect '
                  'the log output for errors.')
    return success


def pin_entry(args, signer):
    pinentry = Pinentry(
        pinentry_path=args.gpg_pinentry_path,
        fallback_to_getpass=True)
    passphrase = pinentry.ask(
        prompt='Enter your gpg passphrase --> ',
        description='Repoman requires a passphrase to unlock key "%s"' %
        signer)
    pinentry.close()
    return passphrase


def get_passphrases(args, repodb):
    passphrases = []
    gpg = repodb._get_gpg(args.gpg_home)
    for idx, signer in enumerate(args.gpg_signer):
        count = 0
        while True:
            try:
                with gpg.key(signer) as sec:
                    if sec.is_unlocked:
                        LOG.warning('gpg key for %s is not locked!', signer)
                        passphrases.append(None)
                        break
                    if args.gpg_passphrase:
                        try:
                            passphrase = args.gpg_passphrase[idx]
                            LOG.warning(
                                'Using passphrase from config for key %s',
                                signer)
                        except IndexError:
                            passphrase = pin_entry(args, signer)
                    else:
                        passphrase = pin_entry(args, signer)
                    with sec.unlock(passphrase):
                        sec.sign('test')
                        passphrases.append(passphrase)
                        break
            except KeyError as ex:
                LOG.fatal('gpg key id "%s" not found in keyring at %s',
                          signer, args.gpg_home)
                sys.exit(1)
            except PGPDecryptionError as ex:
                LOG.error('Could not decrypt gpg key for %s: %s', signer, ex)
                # if they're using pinentry-curses, this error might flash
                # by too quickly to see...
                time.sleep(2)
                if count >= 2:
                    raise
                else:
                    count += 1
            except Exception as ex:
                # something went terribly wrong
                LOG.exception(ex)
                LOG.fatal('something went terribly wrong?!')
                sys.exit(1)
    return passphrases


def publish(args, repodb, repo):
    LOG.info('publishing repository to S3')
    gpg_passphrases = []
    if 'gpg_signer' in args and args.gpg_signer:
        gpg_passphrases = get_passphrases(args, repodb)
    else:
        LOG.warning(
            'no gpg signers present; this will be an insecure apt release')

    return repodb.publish(
        repo,
        dists=args.distribution,
        gpg_home=args.gpg_home,
        gpg_signers=args.gpg_signer,
        gpg_passphrases=gpg_passphrases
    )


def setup(args, repodb, repo):
    """Initialize basic configuration"""
    LOG.info('Setting up repoman!')
    try:
        repodb.initialize(
            dists=args.distribution,
            comps=args.component,
            archs=args.architecture,
            topic_name=args.sns_topic,
            origin=args.origin,
            label=args.label
        )
    except Exception as ex:
        LOG.exception(ex)
        LOG.fatal('Could not initialize simpledb domain: %s', ex)
        LOG.fatal('Failed repoman setup!')
        sys.exit(1)
    try:
        repo.initialize(args.s3_acl, args.s3_region, args.enable_website)
    except Exception as ex:
        LOG.exception(ex)
        LOG.fatal('Could not create S3 bucket: %s', ex)
        LOG.fatal('Failed repoman setup!')
        sys.exit(1)
    repo_print_config(repodb, repo)
    return 0


def checkup(args, repodb, repo):
    retval = 0
    LOG.info('Doing repoman system health check: set skip-checkup to suppress')
    repo_print_config(repodb, repo)
    LOG.info('checking whether simpledb domain exists')
    if not repodb.domain_exists:
        LOG.fatal('Repository simpledb domain does not appear to exist; did '
                  'you forget to run setup?')
        retval = 1
    LOG.info('checking simpledb domain accessibility')
    try:
        result = repodb.add_meta(test_data='foo')
        if not result or result['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Exception('Got error writing to simpledb: %s' % result)
        result = repodb.rm_meta(test_data=True)
        if not result or result['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise Exception('Got error deleting from simpledb: %s' % result)
    except Exception as ex:
        LOG.fatal('Could not successfully perform a test update against '
                  'simpledb domain %s, check your IAM permissions: %s',
                  repodb.domain_name, ex)
        retval = 1
    LOG.info('checking simpledb repo configuration')
    for name, values in iteritems(repodb.meta):
        if name not in ('topic_name', 'test_data'):
            if len(values) < 1:
                LOG.error('Repository DB has no %ss configured, '
                          'did you forget to run setup?', name)
                retval = 1
    LOG.info('checking S3 bucket writability: s3://%s', repo.bucket_name)
    try:
        repo.set_key_from_string('test_key', 'test')
    except Exception as ex:
        LOG.fatal('Could not successfully write to s3://%s/test_key, '
                  'check your IAM permissions: %s', repo.bucket_name, ex)
        retval = 1
    if repodb.topic_name:
        LOG.info('Checking SNS topic "%s" exists', repodb.topic_name)
        topic_arn = repodb.topic_arn
        if not topic_arn:
            LOG.fatal('SNS topic name "%s" does not exist and I do not '
                      'have permissions to create it!', repodb.topic_name)
            retval = 1
    if retval > 0:
        LOG.fatal('Repoman system health check FAILED; see above.')
    else:
        LOG.info('Repomand system health check PASSES: all systems go!')
    return retval


def create_table_data(results, headers):
    table = []
    for name in sorted(results):
        for dist in sorted(results[name]):
            for comp in sorted(results[name][dist]):
                for arch in sorted(results[name][dist][comp]):
                    pkgs = results[name][dist][comp][arch]
                    pkgs.sort(
                        key=lambda x: Dpkg.compare_versions_key(x['version']))
                    for pkg in pkgs:
                        table.append([pkg[x] for x in headers])
    return table


def validate_meta(args, repodb):
    """Preflight check: make sure any distributions, components or
    architectures specified in command line flags are ones that we
    are set up to serve."""
    ret = True
    if 'distribution' in args and args.distribution:
        bad_dists = repodb.find_invalid_metadata(args.distribution, 'dists')
        if bad_dists:
            LOG.error(' "%s" is/are not distributions that this repo '
                      'currently publishes', ','.join(bad_dists))
            ret = False
    if 'component' in args and args.component:
        bad_comps = repodb.find_invalid_metadata(args.component, 'comps')
        if bad_comps:
            LOG.error(' "%s" is/are not components that this repo '
                      'currently publishes', ','.join(bad_comps))
            ret = False
    if 'architecture' in args and args.architecture:
        bad_archs = repodb.find_invalid_metadata(args.architecture, 'archs')
        if bad_archs:
            LOG.error(' "%s" is/are not architectures that this repo '
                      'currently publishes', ','.join(bad_archs))
            ret = False
    return ret


def dump_packages(packages, repodb):
    for name, dists in iteritems(packages):
        for dist, comps in iteritems(dists):
            for comp, archs in iteritems(comps):
                for arch, items in iteritems(archs):
                    for item in items:
                        if 'files' in item:
                            msg = repodb._create_src_msg_from_item(item, dist)
                        else:
                            msg = repodb._create_pkg_msg_from_item(item, dist)
                        print(msg)


def query(args, repodb, repo):
    LOG.debug('doing package query')
    if args.query_hidden:
        # only if --query-hidden should we allow people
        # to pass in empty values here
        LOG.warning('--query-hidden: potentially returning packages that '
                    'are part of deleted distributions, components or '
                    'architectures')
        dists = args.distribution
        comps = args.component
        archs = args.architecture
    else:
        if not validate_meta(args, repodb):
            return 1
        dists = args.distribution or repodb.dists
        comps = args.component or repodb.comps
        archs = args.architecture or repodb.archs

    LOG.debug('querying simpledb')
    results = repodb.query(
        name_wildcard=args.wildcard,
        dists=dists,
        comps=comps,
        archs=archs,
        names=args.package,
        versions=args.version,
        latest_versions=args.latest_versions or 0)

    if not results.keys():
        LOG.fatal('No packages found')
        return False

    if args.outputfmt == 'json':
        print(json.dumps(results, indent=2))
    elif args.outputfmt == 'jsonc':
        print(json.dumps(results))
    elif args.outputfmt == 'packages':
        dump_packages(results, repodb)
    else:
        table = create_table_data(results, HEADERS)
        print('\n' + tabulate(table, headers=HEADERS, tablefmt=args.outputfmt))
    return 0


def confirm(args, evil=[]):
    if not evil and not args.confirm:
        LOG.warning(color(
            'You used the --confirm flag so we are proceeding '
            'without a review/confimation step. Good luck. ', fg='yellow'))
        return True
    if evil:
        for idx, message in enumerate(evil, start=1):
            LOG.warn(color(
                'Found something scary. Problem #%d:\n',
                fg='red', style='bold'), idx)
            sys.stderr.write(
                color(message, fg='black', bg='yellow', style='bold') + '\n\n')
        if args.i_fear_no_evil:
            LOG.warning(color(
                'You used the --i-fear-no-evil flag, so we are '
                'proceeding without review/confirmation even though '
                'you\'re trying to do something scary.  I hope your '
                'insurance is fully paid up...', fg='red'))
        else:
            x = 'I FEAR NO EVIL'
            while input(color(
                    '\nType exactly "%s" to proceed --> ' % x, fg='red')) != x:
                pass
    # yes, we do this even if we pass through the evil check;
    # that's intentional
    if args.confirm:
        while input(color('\nType "c" to confirm --> ', fg='cyan')) != 'c':
            pass
    return True


def cp_prompt(args, candidates, targets, evil=''):
    table = []
    headers = ['name', 'version', 'architecture', 'src dist', 'src comp',
               'dst dist', 'dst comp']
    for name, dists in iteritems(candidates):
        for dist, comps in iteritems(dists):
            for comp, archs in iteritems(comps):
                for arch, pkgs in iteritems(archs):
                    if not pkgs:
                        LOG.warning('no copy actions possible for %s=>%s=>%s',
                                    name, dist, comp)
                        continue
                    for count, pkg in enumerate(pkgs):
                        dst_dist = targets[
                            name][dist][comp][arch][count]['distribution']
                        dst_comp = targets[
                            name][dist][comp][arch][count]['component']
                        table.append(
                            [name, pkg['version'], pkg['architecture'],
                             pkg['distribution'], pkg['component'],
                             dst_dist, dst_comp])
    if not table:
        LOG.fatal('No copy actions were possible with that set of flags.')
        return False
    if not evil and len(table) > 10:
        evil.append(
            'You are about to copy a large number of packages, are '
            'you sure that\'s what you intend?')
    LOG.warning('We are about to do the following copy actions:')
    print('\n' + tabulate(table, headers=headers) + '\n')
    return confirm(args, evil)


def validate_copy_args(args, repodb):
    """Proper argument positioning for a copy operation is
    slightly fussy; make sure cli users do not shoot themselves
    in the foot."""
    evil = []
    if not validate_meta(args, repodb):
        return 1
    if args.src_distribution == args.dst_distribution and \
            args.src_component == args.dst_component:
        LOG.fatal(color(
            'The source and destination distribution and '
            'component flags cannot both match; you\'d have nothing '
            'to copy.', fg='red'))
        return 1
    if not any((args.dst_distribution, args.dst_component)):
        LOG.fatal(color(
            'You must specify at least one of a destination '
            'component or distribution when copying packages, '
            'otherwise what would be the point?', fg='red'))
        return 1
    if not any((args.package, args.version)):
        evil.append(
            'You are potentially copying up to an entire component\'s '
            'worth of packages!')
    return evil


def cp(args, repodb, repo):
    evil = validate_copy_args(args, repodb)
    if evil == 1:
        # we hit a fatal error already
        return 1
    candidates = repodb.get_candidates(
        names=args.package,
        src_dist=args.src_distribution,
        src_comp=args.src_component,
        archs=args.architecture,
        versions=args.version,
        name_wildcard=args.wildcard,
        latest_versions=args.latest_versions or 0)
    candidates, targets = repodb.get_copy_spec(
        candidates=candidates,
        src_dist=args.src_distribution,
        src_comp=args.src_component,
        dst_dist=args.dst_distribution,
        dst_comp=args.dst_component,
        prune_for_promote=args.promote)
    if cp_prompt(args, candidates, targets, evil):
        repodb.do_copy(candidates, targets, repo,
                       overwrite=args.overwrite,
                       auto_purge=args.auto_purge)
    return 0


def rm(args, repodb, repo):
    if not validate_meta(args, repodb):
        return 1
    evil = []
    if not any((args.package, args.version)) and not args.latest_versions:
        evil.append(
            'You are potentially deleting a full component '
            'or distribution!')
    if not any((args.package, args.component,
                args.distribution, args.version,
                args.latest_versions)):
        evil.append('\n\t'.join((
            '\n\t***      YOU      ***',
            '***      ARE      ***',
            '***  POTENTIALLY  ***',
            '***    DELETING   ***',
            '***     EVERY     ***',
            '***    PACKAGE    ***',
            '***      IN       ***',
            '***     THIS      ***',
            '***    ENTIRE     ***',
            '*** REPOSITORY!!! ***')))

    # stupid argparse doesn't seem to want to let us set a default
    # on a flag that stores an int inside a mutually exclusive group
    # so we have to resort to this nonsense
    if isinstance(args.latest_versions, int):
        if args.latest_versions > 0:
            # the meaning of "latest" is reversed from the sense of how
            # it's used when querying or copying:
            # `repoman rm -p foo --exclude-latest` means  delete all BUT
            # the most recent version
            args.latest_versions = -args.latest_versions
        elif args.latest_versions < 0:
            evil.append(
                'You have specified a NEGATIVE value for --exclude-versions. '
                'This means that you are going to DELETE the %s MOST RECENT '
                'packages matching your filter. Review carefully the list of '
                'packages to be deleted and consider your next move carefully.'
                % -args.latest_versions)
        elif args.latest_versions == 0:
            LOG.warning(color(
                '--exclude-recent=0 technically means the same as '
                'not using the flag at all, but the fact that you '
                'did it this way suggests that you might be a little '
                'confused about what you\'re trying to do so I\'m '
                'going to exclude the most recent versions on your '
                'behalf.', fg='red'))
            args.latest_versions = -1

    targets = repodb.query(
        names=args.package,
        dists=args.distribution,
        comps=args.component,
        versions=args.version,
        name_wildcard=args.wildcard,
        latest_versions=args.latest_versions or 0)

    if not targets.keys():
        LOG.warning('No packages to delete; try adjusting your filters.')
        return 0

    table = create_table_data(targets, HEADERS)
    if len(table) > 9:
        evil.append(
            '*** You are about to delete %d packages, which '
            'seems like a lot. Are you sure?' % len(table))
    if not table:
        LOG.info('Found no packages to delete')
        return 0

    print('\n' + tabulate(
        table, headers=HEADERS, tablefmt=args.outputfmt) + '\n')

    LOG.warning('Total packages to be deleted: %s',
                color(str(len(table)), fg='red', style='bold'))
    if confirm(args, evil):
        repodb.do_rm(targets)
    return 0


def backup(args, repodb, repo):
    # this is just like a query with no options, only we
    # also include the _meta key
    backup = {}
    backup['metadata'] = repodb._get_attributes('meta', always_list=True)
    backup['packages'] = repodb.query()
    print(json.dumps(backup, indent=2))
    return 0


def restore(args, repodb, repo):
    with open(args.filename[0]) as fp:
        js = json.loads(fp.read())
        if sorted(js.keys()) != ['metadata', 'packages']:
            LOG.fatal('Backup file corrupt: must have both '
                      'metadata and packages keys.')
            return 1
        meta = js['metadata']
        packages = js['packages']
        for name, dists in iteritems(packages):
            for dist, comps in iteritems(dists):
                for comp, archs in iteritems(comps):
                    for arch, items in iteritems(archs):
                        for item in items:
                            LOG.info('Restoring item: %s', item)
                            repodb._put_item(item)
        LOG.info('Restoring repo configuration: %s', meta)
        repodb._put_attributes('meta', meta)
    return 0


def main():
    retval = 0
    repoman_config = Config(sys.argv[1:])
    args = repoman_config.args

    if args.log_config:
        fp = open(args.log_config, encoding='utf-8')
    else:
        fp = resource_stream('apt_repoman.resources', 'logconfig.json')
    logging.config.dictConfig(json.loads(fp.read().decode()))
    fp.close()

    command = args.command
    if args.debug:
        logging.getLogger('apt_repoman').setLevel(logging.DEBUG)

    connection = Connection(role_arn=args.aws_role)
    repodb = Repodb(args.simpledb_domain, connection=connection)
    repo = Repo(args.s3_bucket, connection=connection)

    funcs = globals()

    if command == 'checkup' or not args.skip_checkup:
        # don't do checkup before setup, that's silly
        if command != 'setup':
            retval += funcs['checkup'](args, repodb, repo)

    if command in ('setup', 'repo', 'add', 'cp', 'rm', 'query',
                   'backup', 'restore'):
        retval += funcs[command](args, repodb, repo)

    # never auto-publish if something above threw an error
    if retval == 0:
        if command == 'publish' or (
                'publish' in args and args.publish is True):
            retval += funcs['publish'](args, repodb, repo)

    return retval

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python

import unittest

from apt_repoman.repo import Repo


class RepoTest(unittest.TestCase):
    def setUp(self):
        self.repo = Repo('testbucket')

    def testGetPkgPathname(self):
        self.assertEqual(
            self.repo._get_pkg_pathname('foo', 'bar', 'baz'),
            'pool/baz/f/foo/bar')


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(RepoTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




# stdlib imports
import logging
import mimetypes
import os

# pypi imports
from botocore.exceptions import ClientError

# internal imports
from apt_repoman.connection import Connection

DEFAULT_ACL = 'bucket-owner-full-control'


LOG = logging.getLogger(__name__)


class RepoError(Exception):
    pass


class BucketNotFoundError(RepoError):
    pass


class KeyNotFoundError(RepoError):
    pass


class KeyExistsError(RepoError):
    pass


class FileNotFoundError(RepoError):
    pass


class Repo(object):
    """Object encapsulating actions on the S3 arm of a repoman repository."""
    def __init__(self, bucket_name, role_arn=None, connection=None):
        self.bucket_name = bucket_name
        self.role_arn = role_arn
        self._log = LOG or logging.getLogger(__name__)
        self._connection = connection or None
        self._s3 = None
        self._bucket = None

    @property
    def connection(self):
        if self._connection is None:
            self._connection = Connection(logger=self._log,
                                          role_arn=self.role_arn)
        return self._connection

    @property
    def s3(self):
        if not self._s3:
            self._s3 = self.connection.s3
        return self._s3

    @property
    def bucket(self):
        if not self._bucket:
            self._bucket = self.s3.Bucket(self.bucket_name)
            if not self._bucket_exists(self._bucket):
                raise BucketNotFoundError(
                    's3://%s/ does not exist or is not visible to me' %
                    self.bucket_name)
        return self._bucket

    def initialize(self, bucket_acl='private', region=None,
                   enable_website=False):
        try:
            self.bucket
            self._log.warning('bucket s3://%s already exists!',
                              self.bucket_name)
        except BucketNotFoundError:
            self._log.warning('Creating s3 bucket: %s', self.bucket_name)
            if enable_website:
                self._log.warning('Enabling the public-read ACL on s3://%s '
                                  'because you have enabled S3 website '
                                  'hosting.', self.bucket_name)
                bucket_acl = 'public-read'
            args = {'Bucket': self.bucket_name,
                    'ACL': bucket_acl}
            if region:
                args['CreateBucketConfiguration'] = {
                    'LocationConstraint': region}
            else:
                self._log.warning('no region specified; creating s3 bucket '
                                  'in US/Standard (us-east-1) region')
            try:
                bucket = self.s3.create_bucket(**args)
                if enable_website:
                    self._log.warning(
                        'Enabling a public website for your bucket')
                    website = bucket.Website()
                    website.put(
                        WebsiteConfiguration={
                            'ErrorDocument': {'Key': 'errors/404.html'},
                            'IndexDocument': {'Suffix': 'index.html'}})
                    policy = bucket.Policy()
                    policy.put(Policy='{"Version":"2012-10-17","Statement":['
                               '{"Effect":"Allow","Principal":"*",'
                               '"Action":["s3:GetObject"],'
                               '"Resource":["arn:aws:s3:::%s/*"]}]}' %
                               self.bucket_name)
                    self._log.info('Bucket website created: '
                                   'http://%s.s3-website-%s.amazonaws.com/',
                                   self.bucket_name, region or 'us-east-1')
                return bucket
            except ClientError as ex:
                if ex.response['Error']['Code'] == 'BucketAlreadyExists':
                    self._log.fatal(
                        'Bucket name "%s" appears to be in use by '
                        'another AWS account. We recommend using DNS '
                        'or classpath-style bucket names: '
                        '"apt.example.com" or "com.example.apt"',
                        self.bucket_name)
                else:
                    self._log.fatal(
                        'Failed to create S3 bucket %s: %s',
                        self.bucket_name, ex.response['Error']['Message'])
                raise
            except Exception as ex:
                self._log.fatal('Failed to create S3 bucket %s: %s',
                                self.bucket_name, ex)
                raise

    def _get_key(self, key_name, s3_obj=None):
        if s3_obj is None:
            return self.s3.Object(self.bucket_name, key_name)
        else:
            return s3_obj(self.bucket_name, key_name)

    def _bucket_exists(self, bucket_obj):
        try:
            bucket_obj.load()
            self._log.debug('s3//%s/ exists!', self.bucket_name)
        except ClientError as ex:
            self._log.error(
                's3//%s/ does not exist or is not visible to me: %s',
                self.bucket_name, ex)
            return False
        return True

    def _key_exists(self, key_obj):
        try:
            metadata = key_obj.metadata
            self._log.debug('key "s3://%s/%s" exists: %s',
                            key_obj.bucket_name, key_obj.key, metadata)
        except ClientError as ex:
            self._log.debug('key "s3://%s/%s" does not exist: %s',
                            key_obj.bucket_name, key_obj.key, ex)
            return False
        return True

    def _set_key_from_file(self, key_name, file_name, acl=DEFAULT_ACL,
                           overwrite=False):
        if not os.path.isfile(file_name):
            raise FileNotFoundError('"%s" is not a file', file_name)
        k = self._get_key(key_name)
        if self._key_exists(k) and not overwrite:
            raise KeyExistsError(
                's3://%s/%s" already exists' % (k.bucket_name, k.key))
        extra_args = {'ACL': acl}
        content_type, encoding = mimetypes.guess_type(file_name)
        if content_type:
            extra_args['ContentType'] = content_type
        if encoding:
            extra_args['ContentEncoding'] = encoding
        self._log.debug('uploading file "%s" with args %s to s3://%s/%s',
                        file_name, extra_args, k.bucket_name, k.key)
        result = k.upload_file(Filename=file_name, ExtraArgs=extra_args)
        self._log.debug('uploaded file "%s" to s3://%s/%s: %s',
                        file_name, k.bucket_name, k.key, result)
        k.reload()
        return k

    def copy_key(self, old_path, new_path, overwrite=False):
        old_key = self._get_key(old_path)
        new_key = self._get_key(new_path)
        if not self._key_exists(old_key):
            raise KeyNotFoundError(
                's3://%s/%s does not exist' % (
                    old_key.bucket_name, old_key.key))
        if not overwrite:
            if self._key_exists(new_key):
                raise KeyExistsError(
                    's3://%s/%s exists and overwrite=False' % (
                        new_key.bucket_name, new_key.key))
        self._log.debug(
            'copying s3://%s/%s to s3://%s/%s', old_key.bucket_name,
            old_key.key, new_key.bucket_name, new_key.key)
        new_key.copy(CopySource={'Bucket': old_key.bucket_name,
                                 'Key': old_key.key})
        new_key.reload()
        return new_key

    def _get_pkg_pathname(self, pkg, filename, dist):
        """Return the relative S3 path name for our package key.
        :param pkg string
        :param filename string
        :param dist string
        """
        return os.path.join('pool', dist, pkg[0], pkg, filename)

    def set_key_from_string(self, key_name, contents,
                            overwrite=True, acl=DEFAULT_ACL,
                            new_sess=True):
        key_url = 's3://{0}/{1}'.format(self.bucket_name, key_name)
        self._log.debug(
            'setting contents of S3 key %s to "%s"', key_url, contents)
        self._log.info('setting contents of %s', key_url)
        k = self._get_key(key_name)
        if not overwrite and self._key_exists(k):
            raise KeyExistsError('%s" already exists' % key_url)
        result = k.put(Body=contents)
        self._log.debug('set contents of %s: %s', key_url,
                        result['ResponseMetadata']['HTTPStatusCode'])
        k.reload()
        return k

    def add_package(self, pkg, dists=[], overwrite=False):
        pkg_name = pkg.get_header('package')
        pkg_file = os.path.basename(pkg.filename)
        for dist in dists:
            key_name = self._get_pkg_pathname(pkg_name, pkg_file, dist)
            try:
                self._log.info('Attempting to upload s3://%s/%s',
                               self.bucket_name, key_name)
                self._set_key_from_file(key_name, pkg.filename,
                                        overwrite=overwrite)
            except KeyExistsError:
                self._log.error(
                    'Key s3://%s/%s already exists, you probably meant '
                    'to copy it within the repo.',
                    self.bucket_name, key_name)
                raise

    def add_source(self, dsc, dists=[], overwrite=False):
        """Iterate over the files proprty of a pydpkg.Dsc object,
        derive the correct S3 pathname for each file and upload
        the file to S3.

        :param dsc: pydpkg.Dsc
        :param dists: list of strings
        :param overwrite: bool
        :returns None
        """
        # blow up immediately if the Dsc is not fully valid:
        dsc.validate()
        name = dsc.source
        for dist in dists:
            upload = dsc.source_files
            self._log.debug('files to upload: %s', upload)
            for src_file in upload:
                key_name = self._get_pkg_pathname(
                    name, os.path.basename(src_file), dist)
                try:
                    self._log.info('Attempting to upload s3://%s/%s',
                                   self.bucket_name, key_name)
                    self._set_key_from_file(key_name, src_file,
                                            overwrite=overwrite)
                except KeyExistsError:
                    self._log.error(
                        'Key s3://%s/%s already exists, you probably meant '
                        'to copy it within the repo.',
                        self.bucket_name, key_name)
                    raise



#!/usr/bin/env python

import pytest
import os
from mock import MagicMock
import json

from apt_repoman import cli
from apt_repoman.repo import Repo
from apt_repoman.repodb import Repodb

import botocore.session
from botocore.stub import Stubber, ANY


@pytest.fixture
def setup():
    s3 = botocore.session.get_session().create_client('s3', region_name='us-east-1')
    sdb = botocore.session.get_session().create_client('sdb', region_name='us-east-1')
    sns = botocore.session.get_session().create_client('sns', region_name='us-east-1')
    repo = Repo('testbucket')
    repodb = Repodb('testdomain')
    repodb._s3 = s3
    repodb._sdb = sdb
    repodb._sns = sns
    repodb._archs = ['amd64', 'i386']
    repodb._dists = ['xenial', 'jessie']
    repodb._comps = ['main', 'nightly']
    repodb._topic_arn = ''
    args = MagicMock()
    _dir = os.path.dirname(__file__)
    return s3, sdb, sns, repo, repodb, args, _dir


def test_query(setup, capsys):
    s3, sdb, sns, repo, repodb, args, _dir = setup
    with open(os.path.join(_dir, 'query_output.json')) as fp:
        select_response = json.loads(fp.read())
    with open(os.path.join(_dir, 'get_attr_meta.json')) as fp:
        get_attr_meta_response = json.loads(fp.read())
    args.architecture = []
    args.distribution = []
    args.component = []
    args.package = []
    args.version = []
    args.query_hidden = False
    args.outputfmt = 'plain'
    args.latest_versions = 0
    with Stubber(sdb) as stub:
        select_params = {
            'ConsistentRead': True,
            'SelectExpression': "select * from `testdomain` where `name` is not null and every(distribution) in ('jessie','xenial') and every(component) in ('main','nightly') and every(architecture) in ('all','amd64','i386','source')"}
        get_attr_params = {'AttributeNames': [], 'ConsistentRead': True, 'DomainName': 'testdomain', 'ItemName': 'meta'}
        stub.add_response('get_attributes', get_attr_meta_response, get_attr_params)
        stub.add_response('select', select_response, select_params)
        expected = open(os.path.join(_dir, 'query_output.txt')).read()
        result = cli.query(args, repodb, repo)
        assert result == 0
        out, err = capsys.readouterr()
        assert out == expected




# stdlib imports
import logging
import time

# pypi imports
from boto3 import Session


LOG = logging.getLogger(__name__)


class Connection(object):

    def __init__(self, role_arn='', profile_name=''):
        self._log = LOG or logging.getLogger(__name__)
        self.role_arn = role_arn
        self.profile_name = profile_name
        self._s3 = None
        self._sdb = None
        self._sts = None
        self._iam = None
        self._sns = None
        self._session = None
        self._caller_id = None

    @property
    def session(self):
        '''Set our object's self._session attribute to a boto3
        session object.  If profile_name is set, use it to pull a
        specific credentials profile from ~/.aws/credentials,
        otherwise use the default credentials path.
        If role_arn is set, use the first session object to
        assume the role, and then overwrite self._session with
        a new session object created using the role credentials.'''
        if self._session is None:
            self._session = self.get_session()
        return self._session

    @property
    def s3(self):
        if self._s3 is None:
            self._s3 = self.get_resource('s3')
        return self._s3

    @property
    def sdb(self):
        if self._sdb is None:
            self._sdb = self.get_client('sdb')
        return self._sdb

    @property
    def sts(self):
        if self._sts is None:
            self._sts = self.get_client('sts')
        return self._sts

    @property
    def iam(self):
        if self._iam is None:
            self._iam = self.get_client('iam')
        return self._iam

    @property
    def sns(self):
        if self._sns is None:
            self._sns = self.get_client('sns')
        return self._sns

    @property
    def caller_id(self):
        if self._caller_id is None:
            self._caller_id = self.sts.get_caller_identity()['Arn']
        return self._caller_id

    def get_session(self):
        if self.profile_name:
            self._log.info(
                'using AWS credential profile %s', self.profile_name)
            try:
                session = Session(profile_name=self.profile_name)
            except Exception as ex:
                self._log.fatal(
                    'Could not connect to AWS using profile %s: %s',
                    self.profile_name, ex)
                raise
        else:
            self._log.debug(
                'getting an AWS session with the default provider')
            session = Session()
        if self.role_arn:
            self._log.info(
                'attempting to assume STS self.role %s', self.role_arn)
            try:
                self.role_creds = session.client('sts').assume_role(
                    RoleArn=self.role_arn,
                    RoleSessionName='repoman-%s' % time.time(),
                    DurationSeconds=3600)['Credentials']
            except Exception as ex:
                self._log.fatal(
                    'Could not assume self.role %s: %s',
                    self.role_arn, ex)
                raise
            session = Session(
                aws_access_key_id=self.role_creds['AccessKeyId'],
                aws_secret_access_key=self.role_creds['SecretAccessKey'],
                aws_session_token=self.role_creds['SessionToken'])
        return session

    def get_client(self, service_name):
        return self.session.client(service_name)

    def get_resource(self, service_name):
        return self.session.resource(service_name)



#!/usr/bin/env python

from setuptools import setup, find_packages

__name__ = 'apt-repoman'
__version__ = '1.0.3'

setup(
    name=__name__,
    packages=find_packages(),
    version=__version__,
    description='A high performance Debian APT repository based on Amazon Web Services',
    author='Nathan J. Mehl',
    author_email='n@climate.com',
    url='https://github.com/theclimatecorporation/repoman',
    download_url='https://github.com/theclimatecorporation/repoman/tarball/%s' % __version__,
    keywords=['apt', 'debian', 'dpkg', 'packaging'],
    package_data={'': ['*.json']},
    install_requires=[
        'PGPy==0.4.1',
        'ansicolors==1.1.8',
        'boto3==1.4.4',
        'configargparse==0.12.0',
        'pydpkg==1.3.1',
        'pysectools==0.4.2',
        'tabulate==0.7.7'
    ],
    extras_require={
        'test': ['mock==2.0.0', 'pep8==1.7.0', 'pytest==3.1.1', 'pylint==1.7.1']
    },
    #scripts=[
    #    'scripts/repoman'
    #],
    entry_points = {
        'console_scripts': ['repoman-cli=apt_repoman.cli:main'],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: System :: Archiving :: Packaging",
        ]
)
